"""
"""
import h5py
import numpy as np

from arrakis_nd.utils.utils import profiler
from arrakis_nd.plugins.plugin import Plugin
from arrakis_nd.utils.track_utils import fit_track
from arrakis_nd.dataset.common import (
    SubProcessType,
    Topology, Physics,
    Blip
)
from arrakis_nd.arrakis.common import blip_data_type


class NeutronPlugin(Plugin):
    """_summary_

    Args:
        Plugin (_type_): _description_
    """
    def __init__(
        self,
        config: dict = {}
    ):
        """
        """
        super(NeutronPlugin, self).__init__(config)

        self.input_products = [
            'daughters',
            'track_id_hit_map',
            'track_id_hit_segment_map',
            'track_id_hit_t0_map',
            'parent_pdg_id',
        ]
        self.output_products = ['tracklette', 'track', 'blip']

        self.neutron_labels = {
            'topology': Topology.Blip.value,
            'physics': Physics.NuclearRecoil.value
        }

    @profiler
    def process_event(
        self,
        event: int,
        flow_file: h5py.File,
        arrakis_file: h5py.File,
        event_indices: dict,
        event_products: dict,
    ):
        """
        """
        trajectories = flow_file['mc_truth/trajectories/data'][event_indices['trajectories']]
        charge = flow_file['charge/calib_final_hits/data'][event_indices['charge']]
        arrakis_charge = arrakis_file['charge_segment/calib_final_hits/data'][event_indices['charge']]
        track_id_hit_map = event_products['track_id_hit_map']
        track_id_hit_segment_map = event_products['track_id_hit_segment_map']
        track_id_hit_t0_map = event_products['track_id_hit_t0_map']
        ancestor_traj_id_map = event_products['ancestor_traj_id_map']
        daughters = event_products['daughters']

        trajectories_traj_ids = trajectories['traj_id']
        trajectories_vertex_ids = trajectories['vertex_id']
        trajectories_parent_ids = trajectories['parent_id']
        trajectories_pdg_ids = trajectories['pdg_id']
        trajectories_xyz_start = trajectories['xyz_start']
        trajectories_xyz_end = trajectories['xyz_end']
        trajectories_pxyz_start = trajectories['pxyz_start']
        trajectories_pxyz_end = trajectories['pxyz_end']
        trajectories_E = trajectories['E_start']
        trajectories_start_subprocess = trajectories['start_subprocess']
        trajectories_end_subprocess = trajectories['end_subprocess']
        charge_x = charge['x']
        charge_y = charge['y']
        charge_z = charge['z']
        charge_E = charge['E']
        charge_io_group = charge['io_group']

        """Grab the neutron which result in ionization"""
        particle_mask = (
            (abs(trajectories_pdg_ids) == 2112)
        )

        """Iterate over the individual (particle_id, vertex_id, traj_index)"""
        """
        First, we generate blip objects associated to the neutron itself.
        These will include elastic/inelastic scatters, but not neutron
        captures.  Those will be handled in a second loop.
        """
        for ii, (particle_id, vertex_id, traj_index) in enumerate(zip(
            trajectories_traj_ids[particle_mask],
            trajectories_vertex_ids[particle_mask],
            event_indices['trajectories'][particle_mask]
        )):
            """Get indices in charge_segments arrays"""
            particle_hits = track_id_hit_map[(particle_id, vertex_id)]

            """Check if neutron has no hits, go to the next particle"""
            if not any(particle_hits):
                continue

            particle_segments = track_id_hit_segment_map[(particle_id, vertex_id)]
            particle_hit_segments = (
                particle_hits, particle_segments
            )

            """Get the associated t0 values"""
            particle_hit_t0s = track_id_hit_t0_map[(particle_id, vertex_id)]

            """Get ancestor traj_id"""
            ancestor_id = ancestor_traj_id_map[(particle_id, vertex_id)]

            particle_charge_xyz = np.array([
                charge_x[particle_hits],
                charge_y[particle_hits],
                charge_z[particle_hits]
            ]).T
            particle_charge_E = charge_E[particle_hits]
            particle_xyz_start = trajectories_xyz_start[particle_mask][ii]
            particle_xyz_end = trajectories_xyz_end[particle_mask][ii]
            particle_pxyz_start = trajectories_pxyz_start[particle_mask][ii]
            particle_pxyz_end = trajectories_pxyz_end[particle_mask][ii]
            particle_E = trajectories_E[particle_mask][ii]
            """Determine track begin and end points"""
            """
            To do this we find the closest hits to the actual track beginning
            and ending from the trajectories.  If these points end up being the
            same, we skip this track, since there's no way to distinguish
            that point as a track anyways.
            """
            particle_start_distances = np.sqrt(
                ((particle_charge_xyz - particle_xyz_start) ** 2).sum(axis=1)
            )
            particle_end_distances = np.sqrt(
                ((particle_charge_xyz - particle_xyz_end) ** 2).sum(axis=1)
            )

            """############################### Set standard labels ###############################"""
            """Set the event_id"""
            arrakis_charge['event_id'][particle_hits] = event

            """Iterate over standard labels"""
            for label, value in self.neutron_labels.items():
                arrakis_charge[label][particle_hit_segments] = value

            """Set the particle label"""
            arrakis_charge['particle'][particle_hit_segments] = trajectories_pdg_ids[particle_mask][ii]

            """Generate unique label"""
            arrakis_charge['unique_topology'][particle_hit_segments] = traj_index

            particle_io_group = charge_io_group[particle_hits]
            unique_particle_io_group = np.unique(particle_io_group)
            particle_hits = np.array(particle_hits)
            particle_segments = np.array(particle_segments)
            particle_hit_t0s = np.array(particle_hit_t0s)
            for io_group in unique_particle_io_group:
                """Isolate hits from this io_group"""
                io_group_mask = (particle_io_group == io_group)
                if not sum(io_group_mask):
                    continue
                io_group_t0s = particle_hit_t0s[io_group_mask]

                """Find the closest points to track begin/end for this io_group"""
                io_group_start_distances = particle_start_distances[io_group_mask]
                io_group_end_distances = particle_end_distances[io_group_mask]
                blip_start_index = np.argmin(io_group_start_distances)
                blip_end_index = np.argmin(io_group_end_distances)

                """Parameterize the trajectory of this blip using t0"""
                io_group_xyz = particle_charge_xyz[io_group_mask]
                track_data = fit_track(io_group_t0s, io_group_xyz)

                """Now generate the associated blip CAF object"""
                """
                The blip data object has the following entries
                that must be filled (see arrakis.common).
                """
                io_group_xyz = particle_charge_xyz[io_group_mask]

                """Set blip type"""
                blip_type = Blip.NuclearRecoil.value

                """Create the CAF object for this blip"""
                blip_data = np.array([(
                    event,
                    io_group,
                    particle_id,
                    vertex_id,
                    ancestor_id,
                    blip_type,
                    particle_xyz_start,
                    particle_xyz_end,
                    [io_group_xyz[blip_start_index]],
                    [io_group_xyz[blip_end_index]],
                    particle_pxyz_start,
                    particle_pxyz_end,
                    [track_data['track_dir']],
                    [track_data['track_enddir']],
                    sum(particle_charge_E[io_group_mask]),
                    1,
                    track_data['track_len_gcm2'],
                    track_data['track_len_cm'],
                    particle_E,
                    [[traj_index]+[0]*19],
                    [[1]+[0]*19])],
                    dtype=blip_data_type
                )

                """Add the new blip to the CAF objects"""
                event_products['blip'].append(blip_data)

        """Write changes to arrakis_file"""
        arrakis_file['charge_segment/calib_final_hits/data'][event_indices['charge']] = arrakis_charge

        """Grab the neutrons which result in capture"""
        particle_mask = (
            (abs(trajectories_pdg_ids) == 2112) &
            (
                (abs(trajectories_end_subprocess) == SubProcessType.HadronCapture.value) |
                (abs(trajectories_end_subprocess) == SubProcessType.HadronCaptureAtRest.value)
            )
        )

        """
        Now we iterate over the neutron captures.
        """
        for ii, (particle_id, vertex_id, traj_index) in enumerate(zip(
            trajectories_traj_ids[particle_mask],
            trajectories_vertex_ids[particle_mask],
            event_indices['trajectories'][particle_mask]
        )):
            """Grab gamma daughters"""
            daughter_ids = daughters[particle_mask][ii]
            neutron_daughter_traj_ids = trajectories_traj_ids[daughter_ids]
            neutron_daughter_pdg_ids = trajectories_pdg_ids[daughter_ids]
            neutron_daughter_start_subprocess = trajectories_start_subprocess[daughter_ids]

            """Iterate over gammas"""
            for jj, daughter_id in enumerate(neutron_daughter_traj_ids):
                if abs(neutron_daughter_pdg_ids[jj]) != 22:
                    continue
                if (
                    (abs(neutron_daughter_start_subprocess[jj]) != SubProcessType.HadronCapture.value) &
                    (abs(neutron_daughter_start_subprocess[jj]) != SubProcessType.HadronCaptureAtRest.value)
                ):
                    continue

                """Get indices in charge_segments arrays"""
                gamma_hits = track_id_hit_map[(daughter_id, vertex_id)]

                gamma_daughter_traj_ids = trajectories_traj_ids[
                    (trajectories_parent_ids == daughter_id) & (trajectories_vertex_ids == vertex_id)
                ]
                for kk, gamma_daughter in enumerate(gamma_daughter_traj_ids):
                    gamma_daughter_hits = track_id_hit_map[(gamma_daughter, vertex_id)]
                    gamma_hits += gamma_daughter_hits




import atexit, shutil, plotly
from os.path import basename
from pathlib import Path
from dash import Dash, dcc, html
import dash_uploader as du
import plotly.graph_objects as go
import plotly.graph_objects as go

from dash.exceptions import PreventUpdate
from dash_extensions.enrich import Output, DashProxy, Input, State, MultiplexerTransform
from dash.dependencies import Input, Output

# Settings and constants
UPLOAD_FOLDER_ROOT = "cache"

def parse_contents(filename):
    data = h5flow.data.H5FlowDataManager(filename, "r")
    num_events = data["charge/events/data"].shape[0]
    return data, num_events


def create_3d_figure(data, evid):
    fig = go.Figure()
    print("here we go")
    # Select the hits for the current event
    prompthits_ev = data["charge/events", "charge/calib_prompt_hits", evid]
    finalhits_ev = data["charge/events", "charge/calib_final_hits", evid]
    # select the segments (truth) for the current event
    try:
        prompthits_segs = data[
            "charge/events",
            "charge/calib_prompt_hits",
            "charge/packets",
            "mc_truth/segments",  # called segments in minirun4
            evid,
        ]
        sim_version = "minirun4"
        print("Found truth info in minirun4 format")
    except:
        print("No truth info in minirun4 format found")
        try:
            prompthits_segs = data[
                "charge/events",
                "charge/calib_prompt_hits",
                "charge/packets",
                "mc_truth/tracks",  # called tracks in minirun3
                evid,
            ]
            sim_version = "minirun3"
            print("Found truth info in minirun3 format")
        except:
            print("No truth info in minirun3 format found")
            prompthits_segs = None

    # Plot the prompt hits
    print("Plotting prompt hits")
    prompthits_traces = go.Scatter3d(
        x=prompthits_ev.data["x"].flatten(),
        y=(prompthits_ev.data["y"].flatten()),
        z=(prompthits_ev.data["z"].flatten()),
        marker_color=prompthits_ev.data["E"].flatten()
        * 1000,  # convert to MeV from GeV for minirun4, not sure for minirun3
        marker={
            "size": 1.75,
            "opacity": 0.7,
            "colorscale": "cividis",
            "colorbar": {
                "title": "Hit energy [MeV]",
                "titlefont": {"size": 12},
                "tickfont": {"size": 10},
                "thickness": 15,
                "len": 0.5,
                "xanchor": "left",
                "x": 0,
            },
        },
        name="prompt hits",
        mode="markers",
        showlegend=True,
        opacity=0.7,
        customdata=prompthits_ev.data["E"].flatten() * 1000,
        hovertemplate="<b>x:%{x:.3f}</b><br>y:%{y:.3f}<br>z:%{z:.3f}<br>E:%{customdata:.3f}",
    )
    print("Adding prompt hits to figure")
    fig.add_traces(prompthits_traces)

    # Plot the final hits
    print("Plotting final hits")
    finalhits_traces = go.Scatter3d(
        x=finalhits_ev.data["x"].flatten(),
        y=(finalhits_ev.data["y"].flatten()),
        z=(finalhits_ev.data["z"].flatten()),
        marker_color=finalhits_ev.data["E"].flatten() * 1000,
        marker={
            "size": 1.75,
            "opacity": 0.7,
            "colorscale": "Plasma",
            "colorbar": {
                "title": "Hit energy [MeV]",
                "titlefont": {"size": 12},
                "tickfont": {"size": 10},
                "thickness": 15,
                "len": 0.5,
                "xanchor": "left",
                "x": 0,
            },
        },
        name="final hits",
        mode="markers",
        visible="legendonly",
        showlegend=True,
        opacity=0.7,
        customdata=finalhits_ev.data["E"].flatten() * 1000,
        hovertemplate="<b>x:%{x:.3f}</b><br>y:%{y:.3f}<br>z:%{z:.3f}<br>E:%{customdata:.3f}",
    )
    print("Adding final hits to figure")
    fig.add_traces(finalhits_traces)

    print("Plotting segments")
    if prompthits_segs is not None:
        segs_traces = plot_segs(
            prompthits_segs[0, :, 0, 0],
            sim_version=sim_version,
            mode="lines",
            name="edep segments",
            visible="legendonly",
            line_color="red",
            showlegend=True,
        )
        print("Adding segments to figure")
        fig.add_traces(segs_traces)

    # Draw the TPC
    print("Drawing TPC")
    tpc_center, anodes, cathodes = draw_tpc(sim_version)
    light_detectors = draw_light_detectors(data, evid)

    print("Adding TPC to figure")
    fig.add_traces(tpc_center)
    fig.add_traces(anodes)
    fig.add_traces(cathodes)
    print("Adding light detectors to figure")
    fig.add_traces(light_detectors)

    return fig


def plot_segs(segs, sim_version="minirun4", **kwargs):
    def to_list(axis):
        if sim_version == "minirun4":
            nice_array = np.column_stack(
                [segs[f"{axis}_start"], segs[f"{axis}_end"], np.full(len(segs), None)]
            ).flatten()
        if sim_version == "minirun3":
            nice_array = np.column_stack(
                [
                    segs[f"{axis}_start"] * 10,
                    segs[f"{axis}_end"] * 10,
                    np.full(len(segs), None),
                ]
            ).flatten()
        return nice_array

    x, y, z = (to_list(axis) for axis in "xyz")

    trace = go.Scatter3d(x=x, y=y, z=z, **kwargs)

    return trace


def draw_tpc(sim_version="minirun4"):
    anode_xs = np.array([-63.931, -3.069, 3.069, 63.931])
    anode_ys = np.array([-19.8543, 103.8543])  # two ys
    anode_zs = np.array([-64.3163, -2.6837, 2.6837, 64.3163])  # four zs
    if sim_version == "minirun4":  # hit coordinates are in cm
        detector_center = (0, -268, 1300)
        anode_ys = anode_ys - (268 + 42)
        anode_zs = anode_zs + 1300
    if sim_version == "minirun3":  # hit coordinates are in mm
        detector_center = (0, 42 * 10, 0)
        anode_xs = anode_xs * 10
        anode_ys = anode_ys * 10
        anode_zs = anode_zs * 10

    center = go.Scatter3d(
        x=[detector_center[0]],
        y=[detector_center[1]],
        z=[detector_center[2]],
        marker=dict(size=3, color="green", opacity=0.5),
        mode="markers",
        name="tpc center",
    )
    anodes = draw_anode_planes(
        anode_xs, anode_ys, anode_zs, colorscale="ice", showscale=False, opacity=0.1
    )
    cathodes = draw_cathode_planes(
        anode_xs, anode_ys, anode_zs, colorscale="burg", showscale=False, opacity=0.1
    )
    return center, anodes, cathodes


def draw_cathode_planes(x_boundaries, y_boundaries, z_boundaries, **kwargs):
    traces = []
    for i_z in range(int(len(z_boundaries) / 2)):
        for i_x in range(int(len(x_boundaries) / 2)):
            z, y = np.meshgrid(
                np.linspace(z_boundaries[i_z * 2], z_boundaries[i_z * 2 + 1], 2),
                np.linspace(y_boundaries.min(), y_boundaries.max(), 2),
            )
            x = (
                (x_boundaries[i_x * 2] + x_boundaries[i_x * 2 + 1])
                * 0.5
                * np.ones(z.shape)
            )
            traces.append(go.Surface(x=x, y=y, z=z, **kwargs))

    return traces


def draw_anode_planes(x_boundaries, y_boundaries, z_boundaries, **kwargs):
    traces = []
    for i_z in range(int(len(z_boundaries) / 2)):
        for i_x in range(int(len(x_boundaries))):
            z, y = np.meshgrid(
                np.linspace(z_boundaries[i_z * 2], z_boundaries[i_z * 2 + 1], 2),
                np.linspace(y_boundaries.min(), y_boundaries.max(), 2),
            )
            x = x_boundaries[i_x] * np.ones(z.shape)

            traces.append(go.Surface(x=x, y=y, z=z, **kwargs))

    return traces


def draw_light_detectors(data, evid):
    try:
        charge = data["charge/events", evid][["id", "unix_ts"]]
        num_light = data["light/events/data"].shape[0]
        light = data["light/events", slice(0, num_light)][
            ["id", "utime_ms"]
        ]  # we have to try them all, events may not be time ordered
    except:
        print("No light information found, not plotting light detectors")
        return []

    match_light = match_light_to_charge_event(charge, light, evid)

    if match_light is None:
        print(
            f"No light event matches found for charge event {evid}, not plotting light detectors"
        )
        return []

    waveforms_all_detectors = get_waveforms_all_detectors(data, match_light)

    # make a list of the sum of the waveform and the channel index
    integral = np.sum(np.sum(waveforms_all_detectors, axis=2), axis=0)
    max_integral = np.max(integral)
    index = np.arange(0, waveforms_all_detectors.shape[1], 1)

    # plot for each of the 96 channels per tpc the sum of the adc values
    drawn_objects = []
    drawn_objects.extend(plot_light_traps(data, integral, index, max_integral))

    return drawn_objects


def match_light_to_charge_event(charge, light, evid):
    """
    Match the light events to the charge event by looking at proximity in time.
    Use unix time for this, since it should refer to the same time in both readout systems.
    For now we just take all the light within 1s from the charge event time.
    """
    matches = []
    for i in range(len(light)):
        if np.abs(light["utime_ms"][i][0] / 1000 - charge["unix_ts"][0]) < 0.5:
            matches.append([charge["id"][0], light["id"][i]])

    match_light = []
    for i in range(len(matches)):
        if (
            matches[i][0] == evid
        ):  # just checking that we get light for the right charge event
            match_light.append(matches[i][1])
    if len(match_light) == 0:
        match_light = None  # no light for this charge event

    return match_light


def get_waveforms_all_detectors(data, match_light):
    """
    Get the light waveforms for the matched light events.
    """
    light_wvfm = data["/light/wvfm", match_light]

    samples_mod0 = light_wvfm["samples"][:, 0:2, :, :]
    samples_mod1 = light_wvfm["samples"][:, 2:4, :, :]
    samples_mod2 = light_wvfm["samples"][:, 4:6, :, :]
    samples_mod3 = light_wvfm["samples"][:, 6:8, :, :]

    sipm_channels_module0 = np.array(
        [2, 3, 4, 5, 6, 7]
        + [9, 10]
        + [11, 12]
        + [13, 14]
        + [18, 19, 20, 21, 22, 23]
        + [25, 26]
        + [27, 28]
        + [29, 30]
        + [34, 35, 36, 37, 38, 39]
        + [41, 42]
        + [43, 44]
        + [45, 46]
        + [50, 51, 52, 53, 54, 55]
        + [57, 58]
        + [59, 60]
        + [61, 62]
    )

    sipm_channels_modules = np.array(
        [4, 5, 6, 7, 8, 9]
        + [10, 11, 12, 13, 14, 15]
        + [20, 21, 22, 23, 24, 25]
        + [26, 27, 28, 29, 30, 31]
        + [36, 37, 38, 39, 40, 41]
        + [42, 43, 44, 45, 46, 47]
        + [52, 53, 54, 55, 56, 57]
        + [58, 59, 60, 61, 62, 63]
    )
    adcs_mod0 = samples_mod0[:, :, sipm_channels_module0, :]
    adcs_mod1 = samples_mod1[:, :, sipm_channels_modules, :]
    adcs_mod2 = samples_mod2[:, :, sipm_channels_modules, :]
    adcs_mod3 = samples_mod3[:, :, sipm_channels_modules, :]

    all_adcs = np.concatenate((adcs_mod0, adcs_mod1, adcs_mod2, adcs_mod3), axis=1)

    # instead of a (m, 8, 48, 1000) array, we want a (m, 4, 96, 1000) array
    # modules instead of tpcs, and 96 channels per module
    m = len(match_light)
    all_modules = all_adcs.reshape((m, 4, 96, 1000))

    # now we make a full array for all the modules
    # could have been done in one step, but this is easier to read
    all_detector = all_modules.reshape((m, 384, 1000))

    return all_detector


def plot_light_traps(data, n_photons, op_indeces, max_integral):
    """Plot optical detectors"""
    drawn_objects = []
    ys = np.flip(
        np.array(
            [
                -595.43,
                -545.68,
                -490.48,
                -440.73,
                -385.53,
                -335.78,
                -283.65,
                -236.65,
                -178.70,
                -131.70,
                -73.75,
                -26.75,
                25.38,
                75.13,
                130.33,
                180.08,
                235.28,
                285.03,
                337.15,
                384.15,
                442.10,
                489.10,
                547.05,
                594.05,
            ]
        )
        / 10
    )
    light_width = ys[1] - ys[0]

    det_bounds = data["/geometry_info/det_bounds/data"]
    COLORSCALE = plotly.colors.make_colorscale(
        plotly.colors.convert_colors_to_same_type(plotly.colors.sequential.YlOrRd)[0]
    )

    for ix in range(0, det_bounds.shape[0]):
        for ilight, light_y in enumerate(ys):
            for iside in range(2):
                opid = ilight + iside * len(ys) + ix * len(ys) * 2
                if opid not in op_indeces:
                    continue
                xx = np.linspace(det_bounds[ix][0][0][0], det_bounds[ix][0][1][0], 2)
                zz = np.linspace(
                    light_y - light_width / 2 + det_bounds[0][0][1] + 0.25,
                    light_y + light_width / 2 + det_bounds[0][0][1] - 0.25,
                    2,
                )

                xx, zz = np.meshgrid(xx, zz)
                light_color = [
                    [
                        0.0,
                        get_continuous_color(
                            COLORSCALE, intermed=max(0, n_photons[opid]) / max_integral
                        ),
                    ],
                    [
                        1.0,
                        get_continuous_color(
                            COLORSCALE, intermed=max(0, n_photons[opid]) / max_integral
                        ),
                    ],
                ]

                if ix % 2 == 0:
                    flip = 0
                else:
                    flip = -1

                opid_str = f"opid_{opid}"
                light_plane = dict(
                    type="surface", 
                    y=np.full(xx.shape, det_bounds[ix][0][0][iside + flip])/10 -240,
                    x=xx/10,
                    z=zz/10+1300,
                    opacity=0.4,
                    hoverinfo="text",
                    ids=[[opid_str, opid_str], [opid_str, opid_str]],
                    customdata=[[opid_str, opid_str], [opid_str, opid_str]],
                    text=f"Optical detector {opid} waveform integral<br>{n_photons[opid]:.2e}",
                    colorscale=light_color,
                    showlegend=False,
                    showscale=False,
                )

                drawn_objects.append(light_plane)

    return drawn_objects

def plot_waveform(data, evid, opid):
    try:
        charge = data["charge/events", evid][["id", "unix_ts"]]
        num_light = data["light/events/data"].shape[0]
        light = data["light/events", slice(0, num_light)][
            ["id", "utime_ms"]
        ]  # we have to try them all, events may not be time ordered
    except:
        print("No light information found, not plotting light waveform")
        return []

    match_light = match_light_to_charge_event(charge, light, evid)

    if match_light is None:
        print(
            f"No light event matches found for charge event {evid}, not plotting light waveform"
        )
        return []

    fig = go.Figure()
    waveforms_all_detectors = get_waveforms_all_detectors(data, match_light)
    wvfm_opid = waveforms_all_detectors[:, opid, :]
    
    x = np.arange(0, 1000, 1)
    y = np.sum(wvfm_opid, axis=0)
    drawn_objects = go.Scatter(x=x, y=y)
    fig.add_traces(drawn_objects)
    
    fig.update_xaxes(title_text='Time [ticks] (1 ns)')
    fig.update_yaxes(title_text='Adc counts')
    fig.update_layout(title_text=f'Waveform for optical detector {opid}')
    return fig



def get_continuous_color(colorscale, intermed):
    """
    Plotly continuous colorscales assign colors to the range [0, 1]. This function computes the intermediate
    color for any value in that range.

    Plotly doesn't make the colorscales directly accessible in a common format.
    Some are ready to use:

        colorscale = plotly.colors.PLOTLY_SCALES["Greens"]

    Others are just swatches that need to be constructed into a colorscale:

        viridis_colors, scale = plotly.colors.convert_colors_to_same_type(plotly.colors.sequential.Viridis)
        colorscale = plotly.colors.make_colorscale(viridis_colors, scale=scale)

    :param colorscale: A plotly continuous colorscale defined with RGB string colors.
    :param intermed: value in the range [0, 1]
    :return: color in rgb string format
    :rtype: str
    """
    if len(colorscale) < 1:
        raise ValueError("colorscale must have at least one color")

    if intermed <= 0 or len(colorscale) == 1:
        return colorscale[0][1]
    if intermed >= 1:
        return colorscale[-1][1]

    for cutoff, color in colorscale:
        if intermed > cutoff:
            low_cutoff, low_color = cutoff, color
        if intermed <= cutoff:
            high_cutoff, high_color = cutoff, color
            break

    # noinspection PyUnboundLocalVariable
    return plotly.colors.find_intermediate_color(
        lowcolor=low_color,
        highcolor=high_color,
        intermed=((intermed - low_cutoff) / (high_cutoff - low_cutoff)),
        colortype="rgb",
    )

def get_layout():
    return html.Div(
        [
            # Hidden divs to store data
            dcc.Location(id="url"),
            dcc.Store(id="filename", storage_type="local", data=None),
            dcc.Store(id='data-length', data=0),
            # Header
            html.H1(children="2x2 event display", style={"textAlign": "center"}),
            html.Div(children="", id="filename-div", style={"textAlign": "center"}),
            # Upload button
            html.Div(
                du.Upload(
                    id="upload-data-div",
                    text="Upload Flow HDF5 File",
                    max_file_size=10000,
                    chunk_size=5,
                    default_style={
                        "width": "15em",
                        "padding": "0",
                        "margin": "0",
                    },
                    pause_button=True,
                    filetypes=["h5"],
                ),
            ),
            # Event ID input box
            dcc.Input(
                    id="input-evid",
                    type="number",
                    placeholder="0",
                    debounce=True,
                    style={
                        "width": "6em",
                        "display": "inline-block",
                        "margin-right": "0.5em",
                        "margin-left": "0.5em",
                    },
                ),
            # Event ID buttons
            html.Button('Previous Event', id='prev-button', n_clicks=0),
            html.Button('Next Event', id='next-button', n_clicks=0),
            dcc.Store(id='event-id', data=0),
            html.Div(id='evid-div', style={"textAlign": "center"}),
            # Graphs
            html.Div([
                # Existing 3D graph
                html.Div(dcc.Graph(id='3d-graph', style={'height': '70vh', 'width': '50vw'})),

                # New Light waveform graph
                html.Div(dcc.Graph(id="light-waveform", style={'height': '50vh', 'width': '35vw'})),

                
            ], style={'display': 'flex'}),
            # New Another Graph (replace with your actual component)
            html.Div(dcc.Graph(id="another-graph", style={'height': '30vh', 'width': '35vw', 'float': 'right'})),
        ]
    )

def add_callbacks(app): 
    '''
    Callbacks
    '''


    # Callback to handle the upload
    # =============================
    @app.callback(
        [
            Output("filename", "data"),
            Output("filename-div", "children"),
            Output("event-id", "data", allow_duplicate=True),
            Output('data-length', 'data'),
        ],
        [
            Input("upload-data-div", "isCompleted"),
        ],
        [
            State("filename", "data"),
            State("upload-data-div", "fileNames"),
            State("upload-data-div", "upload_id"),
        ],
        prevent_initial_call=True
    )
    def upload_file(is_completed, current_filename, filenames, upload_id):
        """
        Upload HDF5 file to cache. If the upload is completed,
        update the filename. Initialise the event ID to 0.
        """
        if not is_completed:
            raise PreventUpdate

        if filenames is not None:
            if upload_id:
                root_folder = Path(UPLOAD_FOLDER_ROOT) / upload_id
            else:
                root_folder = Path(UPLOAD_FOLDER_ROOT)
            _, num_events = parse_contents(str(root_folder / filenames[0]))
            new_filename = str(root_folder / filenames[0])
            return new_filename, basename(filenames[0]), 0, num_events

        return current_filename, "no file uploaded", 0, 0


    # Callbacks to handle the event ID
    # =================================
    @app.callback(
        Output('event-id', 'data', allow_duplicate=True),
        Input('next-button', 'n_clicks'),
        State('event-id', 'data'),
        State('data-length', 'data'),
        prevent_initial_call=True
    )
    def increment(n, evid, max_value):
        if n > 0:
            new_evid = evid + 1
            if new_evid > max_value:  # wrap around
                return 0
            else:
                return new_evid

    @app.callback(
        Output('event-id', 'data', allow_duplicate=True),
        Input('prev-button', 'n_clicks'),
        State('event-id', 'data'),
        State('data-length', 'data'),
        prevent_initial_call=True
    )
    def decrement(n, evid, max_value):
        if n > 0:
            if evid > 0:
                return evid - 1
            return max_value - 1 # wrap around

    @app.callback(
        Output('event-id', 'data', allow_duplicate=True),
        Input('input-evid', 'value'),
        State('data-length', 'data'),
        prevent_initial_call=True
    )
    def set_evid(value, max_value):
        if value is not None:
            if value > max_value:  # not possible to go higher than max value
                return max_value
            else:
                return value

    @app.callback(
        Output('evid-div', 'children'),
        Input('event-id', 'data'),
        State('data-length', 'data'),
    )
    def update_div(evid, max_value):
        return f'Event ID: {evid}/{max_value}'

    # Callback to display the event
    # =============================
    @app.callback(
        Output('3d-graph', 'figure'),
        Input('filename', 'data'),
        Input('event-id', 'data'),
        prevent_initial_call=True
    )
    def update_graph(filename, evid):
        if filename is not None:
            data, _ = parse_contents(filename)
            return create_3d_figure(data, evid)
        
    @app.callback(
        Input('filename', 'data'),
        Input('event-id', 'data'),
        Input('3d-graph', 'figure'),
        Input('3d-graph', 'clickData'),
        Output('light-waveform', 'figure'),
    )
    def update_light_waveform(filename, evid, graph, click_data):
        if click_data:
            curvenum = int(click_data["points"][0]["curveNumber"])
            print(curvenum) # this is the curve number from the clickdata of the graph
            opid = int(graph['data'][curvenum]['ids'][0][0].split('_')[1])
            print(opid) # the opid related to the curvenumber
            if filename is not None:
                data, _ = parse_contents(filename)
                return plot_waveform(data, evid, opid)
        return go.Figure()


    # Cleaning up
    # ===========
    @atexit.register
    def clean_cache():
        """Delete uploaded files"""
        try:
            shutil.rmtree(Path(UPLOAD_FOLDER_ROOT))
        except OSError as err:
            print("Can't clean %s : %s" % (UPLOAD_FOLDER_ROOT, err.strerror))

def run_my_dash_app():
    # Create the app
    app = DashProxy(__name__, title="2x2 event display")
    du.configure_upload(app, UPLOAD_FOLDER_ROOT)  # without this upload will not work

    # App layout
    app.layout = get_layout()

    # Add the callbacks
    add_callbacks(app)

    # app.run(jupyter_mode="jupyterlab")
    app.run_server(
            jupyter_mode=None,
            jupyter_server_url="http://localhost",
            host="localhost",
            port=8052,
        )


"""
"""
import h5py
import numpy as np
from scipy.interpolate import splprep, splev
import warnings
from scipy.integrate import quad, IntegrationWarning
import fastcluster

from arrakis_nd.utils.utils import profiler, integrand
from arrakis_nd.plugins.plugin import Plugin
from arrakis_nd.utils.track_utils import fit_track
from arrakis_nd.dataset.common import (
    ProcessType, SubProcessType,
    Topology, Physics
)
from arrakis_nd.utils.utils import fiducialized_vertex


class ProtonPlugin(Plugin):
    """

    """
    def __init__(
        self,
        config: dict = {}
    ):
        """
        """
        super(ProtonPlugin, self).__init__(config)

        self.input_products = [
            'daughters',
            'track_id_hit_map',
            'track_id_hit_segment_map',
            'track_id_hit_t0_map'
        ]
        self.output_products = []

    @profiler
    def process_event(
        self,
        event: int,
        flow_file: h5py.File,
        arrakis_file: h5py.File,
        event_indices: dict,
        event_products: dict,
    ):
        """
        """
        trajectories = flow_file['mc_truth/trajectories/data'][event_indices['trajectories']]
        interactions = flow_file['mc_truth/interactions/data'][event_indices['interactions']]
        charge = flow_file['charge/calib_final_hits/data'][event_indices['charge']]
        charge_backtrack = flow_file['mc_truth/calib_final_hit_backtrack/data'][event_indices['charge']]
        track_id_hit_map = event_products['track_id_hit_map']
        track_id_hit_segment_map = event_products['track_id_hit_segment_map']
        track_id_hit_t0_map = event_products['track_id_hit_t0_map']
        charge_back_track_segments = charge_backtrack['segment_id']

        trajectories_traj_ids = trajectories['traj_id']
        trajectories_vertex_ids = trajectories['vertex_id']
        trajectories_parent_ids = trajectories['parent_id']
        trajectories_pdg_ids = trajectories['pdg_id']
        trajectories_xyz_start = trajectories['xyz_start']
        trajectories_xyz_end = trajectories['xyz_end']
        trajectories_pxyz_start = trajectories['pxyz_start']
        trajectories_pxyz_end = trajectories['pxyz_end']
        trajectories_E = trajectories['E_start']
        trajectories_t_start = trajectories['t_start']
        trajectories_start_process = trajectories['start_process']
        trajectories_start_subprocess = trajectories['start_subprocess']
        charge_x = charge['x']
        charge_y = charge['y']
        charge_z = charge['z']
        charge_E = charge['E']
        charge_io_group = charge['io_group']

        """Grab the mips/hips"""
        particle_mask = (
            (abs(trajectories_pdg_ids) == 13) |
            (abs(trajectories_pdg_ids) == 15) |
            (abs(trajectories_pdg_ids) == 211) |
            (abs(trajectories_pdg_ids) == 321) |
            (trajectories_pdg_ids == 2212)
        )
        proton_mask = (
            (trajectories_pdg_ids == 2212) &
            (trajectories_start_subprocess == 121)
        )
        particle_hits = []
        for (particle_id, vertex_id) in zip(
            trajectories_traj_ids[particle_mask],
            trajectories_vertex_ids[particle_mask]
        ):
            particle_hits.append(np.array(track_id_hit_map[(particle_id, vertex_id)], dtype=int))

        particle_hits = np.concatenate(particle_hits, dtype=int)
        if not any(particle_hits):
            return

        particle_charge_xyz = np.array([
            charge_x[np.unique(particle_hits)],
            charge_y[np.unique(particle_hits)],
            charge_z[np.unique(particle_hits)]
        ]).T

        """Iterate over the individual (particle_id, vertex_id, traj_index)"""
        for ii, (proton_id, vertex_id, traj_index) in enumerate(zip(
            trajectories_traj_ids[proton_mask],
            trajectories_vertex_ids[proton_mask],
            event_indices['trajectories'][proton_mask]
        )):
            """Get indices in charge_segments arrays"""
            proton_hits = track_id_hit_map[(proton_id, vertex_id)]

            """Check if mip has no hits, go to the next proton"""
            if not any(proton_hits):
                continue

            proton_segments = track_id_hit_segment_map[(proton_id, vertex_id)]
            proton_hit_segments = (
                proton_hits, proton_segments
            )

            """Get the associated t0 values"""
            proton_hit_t0s = track_id_hit_t0_map[(proton_id, vertex_id)]
            proton_t_start = trajectories_t_start[proton_mask][ii]

            """Get neutrino vertex and check whether it's fiducialized"""
            vert_mask = interactions['vertex_id'] == vertex_id
            nu_vert = interactions[vert_mask]
            vert_loc = nu_vert['vertex'][0]
            neutrino_vertex_fiducialized = fiducialized_vertex(nu_vert['vertex'][0])

            """Get proton parent info"""
            parent_pdg = 0
            parent_E = 0.0
            parent_distance = 0.0
            parent_pxyz_start = [[0, 0, 0]]
            parent_pxyz_end = [[0, 0, 0]]
            proton_parent_dt = 0
            proton_parent_id = trajectories_parent_ids[proton_mask][ii]

            if proton_parent_id != -1:
                proton_parent_index = np.where(
                    (trajectories_traj_ids == proton_parent_id) & (trajectories_vertex_ids == vertex_id)
                )
                proton_parent = trajectories[proton_parent_index]
                parent_pdg = proton_parent['pdg_id']
                parent_E = proton_parent['E_start']
                parent_distance = np.sqrt(
                    (proton_parent['xyz_start'][0][0] - proton_parent['xyz_end'][0][0]) ** 2 +
                    (proton_parent['xyz_start'][0][1] - proton_parent['xyz_end'][0][1]) ** 2 +
                    (proton_parent['xyz_start'][0][2] - proton_parent['xyz_end'][0][2]) ** 2
                )
                parent_pxyz_start = proton_parent['pxyz_start']
                parent_pxyz_end = proton_parent['pxyz_end']
                proton_parent_dt = (proton_t_start - vert_loc[3])

            """Get grandparent info"""
            grandparent_pdg = 0
            if proton_parent_id != -1:
                proton_grandparent_id = proton_parent['parent_id']
                if proton_grandparent_id != -1:
                    proton_grandparent_index = np.where(
                        (trajectories_traj_ids == proton_grandparent_id) & (trajectories_vertex_ids == vertex_id)
                    )
                    grandparent_pdg = trajectories[proton_grandparent_index]['pdg_id']

            """############################### Generate track data ###############################"""
            proton_charge_xyz = np.array([
                charge_x[proton_hits],
                charge_y[proton_hits],
                charge_z[proton_hits]
            ]).T
            proton_charge_E = charge_E[proton_hits]
            proton_xyz_start = trajectories_xyz_start[proton_mask][ii]
            proton_xyz_end = trajectories_xyz_end[proton_mask][ii]
            proton_pxyz_start = trajectories_pxyz_start[proton_mask][ii]
            proton_pxyz_end = trajectories_pxyz_end[proton_mask][ii]
            proton_E = trajectories_E[proton_mask][ii]

            """Check whether proton is fiducialized"""
            proton_xyz_start_fiducialized = fiducialized_vertex(proton_xyz_start)
            proton_xyz_end_fiducialized = fiducialized_vertex(proton_xyz_end)

            """
            To do this we find the closest hits to the actual track beginning
            and ending from the trajectories.  If these points end up being the
            same, we skip this track, since there's no way to distinguish
            that point as a track anyways.
            """
            proton_start_distances = np.sqrt(
                ((proton_charge_xyz - proton_xyz_start) ** 2).sum(axis=1)
            )
            proton_end_distances = np.sqrt(
                ((proton_charge_xyz - proton_xyz_end) ** 2).sum(axis=1)
            )
            closest_start_index = np.argmin(proton_start_distances)
            closest_end_index = np.argmin(proton_end_distances)

            proton_start_index = (
                proton_hits[closest_start_index],
                proton_segments[closest_start_index]
            )
            proton_end_index = (
                proton_hits[closest_end_index],
                proton_segments[closest_end_index]
            )

            """Generate track fit data"""
            track_data = fit_track(proton_hit_t0s, proton_charge_xyz)

            """Determine amount by which hits are shared from things other than the proton"""
            proton_hit_segment_ids = charge_back_track_segments[np.unique(proton_hits)]
            true_proton_hit_purity = float(len(proton_hits)) / float(np.count_nonzero(proton_hit_segment_ids))

            """############################### Generate clusters ###############################"""
            linked = fastcluster.linkage_vector(particle_charge_xyz, method='ward')

            n_clusters = len(linked)
            cluster_mapping = {i: [i] for i in range(len(particle_charge_xyz))}

            # Initialize a dictionary to hold lists of indices for each cluster
            clusters = {i: [] for i in range(-1, n_clusters)}

            for ii, row in enumerate(linked):
                cluster_1, cluster_2 = int(row[0]), int(row[1])
                new_cluster_index = len(particle_charge_xyz) + ii  # New index for the newly formed cluster

                # Combine the points from the clusters being merged and update the mapping
                combined_points = cluster_mapping[cluster_1] + cluster_mapping[cluster_2]
                cluster_mapping[new_cluster_index] = combined_points
                clusters[new_cluster_index] = combined_points

            # Iterate over the linkage matrix
            lowest_completeness_cluster = 0
            lowest_completeness_label = 0
            lowest_completeness = 0.0
            lowest_completeness_purity = 0.0

            """Set up list of proton points"""
            proton_points = np.zeros(len(np.unique(particle_hits)), dtype=bool)
            particle_proton_points = np.where(
                np.isin(np.unique(proton_hits), np.unique(particle_hits))
            )[0]
            proton_points[particle_proton_points] = True

            for ii, (label, combined_points) in enumerate(clusters.items()):
                if label == -1:
                    continue
                particle_points = np.zeros(len(np.unique(particle_hits)), dtype=bool)
                particle_points[combined_points] = True

                if not sum(particle_points):
                    continue

                completeness = sum(particle_points & proton_points) / sum(proton_points)
                purity = sum(particle_points & proton_points) / sum(particle_points)

                if (completeness >= lowest_completeness) and (lowest_completeness < 1.0):
                    lowest_completeness = completeness
                    lowest_completeness_label = label
                    lowest_completeness_cluster = ii
                    lowest_completeness_purity = purity

            """Construct output CAF"""
            nar_inelastic_data_type = np.dtype([
                ('event_id', 'i4'),
                ('vertex_id', 'i4'),
                ('proton_id', 'i4'),
                ('proton_xyz_start', 'f4', (1, 3)),
                ('nu_vertex', 'f4', (1, 3)),
                ('proton_total_energy', 'f4'),
                ('proton_vis_energy', 'f4'),
                ('proton_length', 'f4'),
                ('proton_pxyz_start', 'f4', (1, 3)),
                ('proton_pxyz_end', 'f4', (1, 3)),
                ('parent_total_energy', 'f4'),
                ('parent_length', 'f4'),
                ('parent_pxyz_start', 'f4', (1, 3)),
                ('parent_pxyz_end', 'f4', (1, 3)),
                ('nu_proton_dt', 'f4'),
                ('nu_proton_distance', 'f4'),
                ('parent_pdg', 'i4'),
                ('grandparent_pdg', 'i4'),
                ('primary_pdg', 'i4'),
                ('primary_length', 'f4'),
                ('neutrino_vertex_fiducialized', 'i4'),
                ('proton_xyz_start_fiducialized', 'i4'),
                ('proton_xyz_end_fiducialized', 'i4'),
                ('truth_segment_overlap', 'f4'),
                ('best_completeness_cluster', 'i4'),
                ('best_completeness', 'f4'),
                ('best_purity', 'f4')
            ])
            nar_inelastic_data = np.array([(
                event,
                vertex_id,
                proton_id,
                proton_xyz_start,
                vert_loc[:3],
                proton_E,
                sum(proton_charge_E),
                np.sqrt(
                    (proton_xyz_start[0] - proton_xyz_end[0]) ** 2 +
                    (proton_xyz_start[1] - proton_xyz_end[1]) ** 2 +
                    (proton_xyz_start[2] - proton_xyz_end[2]) ** 2
                ),
                proton_pxyz_start,
                proton_pxyz_end,
                parent_E,
                parent_distance,
                parent_pxyz_start,
                parent_pxyz_end,
                proton_parent_dt,
                np.sqrt(
                    (proton_xyz_start[0] - vert_loc[0]) ** 2 +
                    (proton_xyz_start[1] - vert_loc[1]) ** 2 +
                    (proton_xyz_start[2] - vert_loc[2]) ** 2
                ),
                parent_pdg,
                grandparent_pdg,
                0,
                0,
                neutrino_vertex_fiducialized,
                proton_xyz_start_fiducialized,
                proton_xyz_end_fiducialized,
                true_proton_hit_purity,
                lowest_completeness_cluster,
                lowest_completeness,
                lowest_completeness_purity)],
                dtype=nar_inelastic_data_type
            )
            event_products['nar_inelastic'].append(nar_inelastic_data)



import plotly.graph_objects as go
from dash import dcc,html
from arrakis_nd.utils.display.utils import *
import h5py
class VisEvent:
    '''
    This class is used to visualize the events in three different plots:

    '''
    def __init__(self,folder,file,event):
        print("Creating a VisEvent object")

        self.folder = folder
        self.file = file
        self.event = event
        self.data = h5py.File(self.folder + file, "r")
        self.sim_version = ''
        

    def create_3d_figure(self):
        fig = go.Figure()
        print("here we go")
        # Select the hits for the current event
        prompthits_ev = self.data["charge/calib_prompt_hits/data"]         
        finalhits_ev = self.data["charge/calib_final_hits/data"]
        # select the segments (truth) for the current event
        try:
            prompthits_segs = self.data["mc_truth/segments"]
            self.sim_version = "minirun4"
            print("Found truth info in minirun4 format")
        except:
            print("No truth info in minirun4 format found")
            try:
                prompthits_segs = self.data["mc_truth/tracks"]
                self.sim_version = "minirun3"
                print("Found truth info in minirun3 format")
            except:
                print("No truth info in minirun3 format found")
                prompthits_segs = None
        # TODO: understand/fix this
                
        # Plot the prompt hits
        print("Plotting prompt hits")
        prompthits_traces = go.Scatter3d(
            x=prompthits_ev["x"].flatten(),
            y=(prompthits_ev["y"].flatten()),
            z=(prompthits_ev["z"].flatten()),
            marker_color=prompthits_ev["E"].flatten()
            * 1000,  # convert to MeV from GeV for minirun4, not sure for minirun3
            marker={
                "size": 1.75,
                "opacity": 0.7,
                "colorscale": "cividis",
                "colorbar": {
                    "title": "Hit energy [MeV]",
                    "titlefont": {"size": 12},
                    "tickfont": {"size": 10},
                    "thickness": 15,
                    "len": 0.5,
                    "xanchor": "left",
                    "x": 0,
                },
            },
            name="prompt hits",
            mode="markers",
            showlegend=True,
            opacity=0.7,
            customdata=prompthits_ev["E"].flatten() * 1000,
            hovertemplate="<b>x:%{x:.3f}</b><br>y:%{y:.3f}<br>z:%{z:.3f}<br>E:%{customdata:.3f}",
        )
        print("Adding prompt hits to figure")
        fig.add_traces(prompthits_traces)

        # Plot the final hits
        print("Plotting final hits")
        finalhits_traces = go.Scatter3d(
            x=finalhits_ev["x"].flatten(),
            y=(finalhits_ev["y"].flatten()),
            z=(finalhits_ev["z"].flatten()),
            marker_color=finalhits_ev["E"].flatten() * 1000,
            marker={
                "size": 1.75,
                "opacity": 0.7,
                "colorscale": "Plasma",
                "colorbar": {
                    "title": "Hit energy [MeV]",
                    "titlefont": {"size": 12},
                    "tickfont": {"size": 10},
                    "thickness": 15,
                    "len": 0.5,
                    "xanchor": "left",
                    "x": 0,
                },
            },
            name="final hits",
            mode="markers",
            visible="legendonly",
            showlegend=True,
            opacity=0.7,
            customdata=finalhits_ev["E"].flatten() * 1000,
            hovertemplate="<b>x:%{x:.3f}</b><br>y:%{y:.3f}<br>z:%{z:.3f}<br>E:%{customdata:.3f}",
        )
        print("Adding final hits to figure")
        fig.add_traces(finalhits_traces)

        # print("Plotting segments")
        # if prompthits_segs is not None:
        #     segs_traces = plot_segs(
        #         # prompthits_segs[0, :, 0, 0],
        #         sim_version=self.sim_version,
        #         mode="lines",
        #         name="edep segments",
        #         visible="legendonly",
        #         line_color="red",
        #         showlegend=True,
        #     )
        #     print("Adding segments to figure")
        #     fig.add_traces(segs_traces)
        # TODO: fix this

        # Draw the TPC
        print("Drawing TPC")
        tpc_center, anodes, cathodes = draw_tpc(self.sim_version)
        light_detectors = draw_light_detectors(self.data, 0)

        print("Adding TPC to figure")
        fig.add_traces(tpc_center)
        fig.add_traces(anodes)
        fig.add_traces(cathodes)
        print("Adding light detectors to figure")
        fig.add_traces(light_detectors)

        return fig

    def get_layout(self):
        figure3d = self.create_3d_figure()
        return html.Div(
            [
                # Hidden divs to store data
                dcc.Location(id="url"),
                dcc.Store(id="filename", storage_type="local", data=None),
                dcc.Store(id='data-length', data=0),
                # Header
                html.H1(children="2x2 event display", style={"textAlign": "center"}),
                # Graphs
                html.Div([
                    # Existing 3D graph
                    html.Div(dcc.Graph(id='3d-graph', style={'height': '70vh', 'width': '50vw'}, figure=figure3d)),
                    # New Light waveform graph
                    html.Div(dcc.Graph(id="light-waveform", style={'height': '50vh', 'width': '35vw'})),
                ], style={'display': 'flex'}),
                # New Another Graph (replace with your actual component)
                html.Div(dcc.Graph(id="another-graph", style={'height': '30vh', 'width': '35vw', 'float': 'right'})),
            ]
        )


"""
"""
import numpy as np

"""
Second, we set up arrays for CAF-like reco objects which consist
of the following types:
    (1) tracklettes - pieces of contiguous track like objects.
        (a) start (x,y,z) - geant4 track start point
        (b) end (x,y,z) - geant4 track end point
        (c) start_hit (x,y,z) - corresponding hit start point
        (d) end_hit (x,y,z) - corresponding hit end point
        (e) dir (u_x,u_y,u_z) - geant4 track momentum start dir
        (f) enddir (u_x,u_y,u_z) - geant4 track momentum end dir
        (g) dir_hit (u_x,u_y,u_z) - estimate of track direction taken from start_hit point
        (h) enddir_hit (u_x,u_y,u_z) - estimate of track direction takend from end_hit point
        (i) Evis - visible energy in voxels corresponding to this track
        (j) qual - reco specific quality metric
        (k) len_gcm2 - track length in g/cm2
        (l) len_cm - track length in centimeter
        (m) E - track energy estimate in MeV
        (n) truth - associated true particle in flow file (if relevant) (i.e. track_id)
        (o) truthOverlap - fractional overlap between this track and true particle
    (2) tracks - collections of tracklettes which define a complete track.
    (3) fragments - pieces of contiguous shower like objects.
        (a) start (x,y,z) - geant4 shower start point
        (b) start_hit (x,y,z) - corresponding hit start point
        (c) dir (u_x,u_y,u_z) - geant4 shower direction
        (d) dir_hit (u_x,u_y,u_z) - estimate of shower direction taken from start_hit point
        (e) Evis - visible energy in voxels corresponding to this shower
        (f) qual - reco specific quality metric
        (g) len_gcm2 - track length in g/cm2
        (h) len_cm - track length in centimeter
        (i) E - track energy estimate in MeV
        (j) truth - associated true particle in flow file (if relevant) (i.e. track_id)
        (k) truthOverlap - fractional overlap between this shower and true particle
    (4) showers - collections of fragments which define a complete shower.
    (5) blips - collections of points associated to blip-like activity.
        (a) start (x,y,z) - geant4 start position of this blip object
        (b) start_hit (x,y,z) - corresponding hit start point
        (c) Evis - visible energy in voxels corresponding to this blip
        (d) E - reconstructed energy (GeV)
        (e) bliphyp - hypothesis for this blip's identity
        (f) truth - associated true particle in flow file (if relevant) (i.e track_id)
        (g) truthOverlap - fractional overlap between this blip and true particle
    (6) particles - particles within an event that are associated to tracks/showers/blips.
        (a) primary - is this reco particle a primary one (i.e. eminates directly from vertex)?
        (b) pdg - pdg code inferred for this particle
        (c) tgtA - atomic number of nucleus this particle was reconstructed in
        (d) score - PID score for this particle
        (e) Evis - visible energy in voxels corresponding to this particle
        (f) E - reconstructed energy (GeV)
        (g) E_method - method used to determine energy for the particle
        (h) p (p_x,p_y,p_z) - reconstructed momentum for this particle
        (i) start (x,y,z) - reconstructed start point of this particle
        (j) end (x,y,z) - reconstructed end point of this particle
        (k) start_hit (x,y,z) - reconstructed start hit of this particle
        (l) end_hit (x,y,z) - reconstructed end hit of this particle
        (m) contained - contained in LAr TPC?
        (n) truth - associated true particle in flow file (if relevant) (i.e. track_id)
        (o) truthOverlap - fractional overlap between this reco particle and true particle
    (7) interactions - collections of tracks/showers/blips which define interactions of interest.
        (a) vtx (x,y,z) - reconstructed vertex location
        (b) dir (u_x,u_y,u_z) - hypothesis for this interaction's parent particle direction
            (b.i) lngtrk - direction using longest track
            (b.ii) heshw - direction using highest energy shower
        (c) nuhyp - hypothesis for this interaction's neutrino identity
            (c.i) isnubar
            (c.ii) nue
            (c.iii) numu
            (c.iv) nutau
            (c.v) nc
            (c.vi) protons0
            (c.vii) protons1
            (c.viii) protons2
            (c.ix) protonsN
            (c.x) chgpi0
            (c.xi) chgpi1
            (c.xii) chgpi2
            (c.xiii) chgpiN
            (c.xiv) pizero0
            (c.xv) pizero1
            (c.xvi) pizero2
            (c.xvii) pizeroN
            (c.xviii) neutron0
            (c.xix) neutron1
            (c.xx) neutron2
            (c.xxi) neutronN
        (d) Enu - hypothesis for this interaction's neutrino energy
        (e) part - collections of reconstructed particles
        (f) truth - indicies of true_interactions in flow file (if relevant)
        (g) truthOverlap - fractional overlap between this reco interaction and each true interaction
"""

tracklette_data_type = np.dtype([
    ('event_id', 'i4'),
    ('tpc_id', 'i4'),
    ('tracklette_id', 'i4'),
    ('vertex_id', 'i8'),
    ('tracklette_type', 'i4'),
    ('start', 'f4', (1, 3)),
    ('end', 'f4', (1, 3)),
    ('start_hit', 'f4', (1, 3)),
    ('end_hit', 'f4', (1, 3)),
    ('dir', 'f4', (1, 3)),
    ('enddir', 'f4', (1, 3)),
    ('dir_hit', 'f4', (1, 3)),
    ('enddir_hit', 'f4', (1, 3)),
    ('Evis', 'f4'),
    ('qual', 'f4'),
    ('len_gcm2', 'f4'),
    ('len_cm', 'f4'),
    ('E', 'f4'),
    ('truth', 'i4', (1, 20)),
    ('truthOverlap', 'f4', (1, 20)),
])

track_data_type = np.dtype([
    ('event_id', 'i4'),
    ('track_id', 'i4'),
    ('vertex_id', 'i8'),
    ('track_type', 'i4'),
    ('start', 'f4', (1, 3)),
    ('end', 'f4', (1, 3)),
    ('start_hit', 'f4', (1, 3)),
    ('end_hit', 'f4', (1, 3)),
    ('dir', 'f4', (1, 3)),
    ('enddir', 'f4', (1, 3)),
    ('dir_hit', 'f4', (1, 3)),
    ('enddir_hit', 'f4', (1, 3)),
    ('Evis', 'f4'),
    ('qual', 'f4'),
    ('len_gcm2', 'f4'),
    ('len_cm', 'f4'),
    ('E', 'f4'),
    ('truth', 'i4', (1, 20)),
    ('truthOverlap', 'f4', (1, 20)),
])

fragment_data_type = np.dtype([
    ('event_id', 'i4'),
    ('tpc_id', 'i4'),
    ('fragment_id', 'i4'),
    ('vertex_id', 'i8'),
    ('ancestor_id', 'i4'),
    ('fragment_type', 'i4'),
    ('start', 'f4', (1, 3)),
    ('end', 'f4', (1, 3)),
    ('start_hit', 'f4', (1, 3)),
    ('end_hit', 'f4', (1, 3)),
    ('dir', 'f4', (1, 3)),
    ('enddir', 'f4', (1, 3)),
    ('dir_hit', 'f4', (1, 3)),
    ('enddir_hit', 'f4', (1, 3)),
    ('Evis', 'f4'),
    ('qual', 'f4'),
    ('len_gcm2', 'f4'),
    ('len_cm', 'f4'),
    ('E', 'f4'),
    ('truth', 'i4', (1, 20)),
    ('truthOverlap', 'f4', (1, 20)),
])

shower_data_type = np.dtype([
    ('event_id', 'i4'),
    ('shower_id', 'i4'),
    ('vertex_id', 'i8'),
    ('shower_type', 'i4'),
    ('start', 'f4', (1, 3)),
    ('end', 'f4', (1, 3)),
    ('start_hit', 'f4', (1, 3)),
    ('end_hit', 'f4', (1, 3)),
    ('dir', 'f4', (1, 3)),
    ('enddir', 'f4', (1, 3)),
    ('dir_hit', 'f4', (1, 3)),
    ('enddir_hit', 'f4', (1, 3)),
    ('Evis', 'f4'),
    ('qual', 'f4'),
    ('len_gcm2', 'f4'),
    ('len_cm', 'f4'),
    ('width', 'f4'),
    ('angle', 'f4'),
    ('E', 'f4'),
    ('truth', 'i4', (1, 20)),
    ('truthOverlap', 'f4', (1, 20)),
])

blip_data_type = np.dtype([
    ('event_id', 'i4'),
    ('tpc_id', 'i4'),
    ('blip_id', 'i4'),
    ('vertex_id', 'i8'),
    ('ancestor_id', 'i4'),
    ('blip_type', 'i4'),
    ('start', 'f4', (1, 3)),
    ('end', 'f4', (1, 3)),
    ('start_hit', 'f4', (1, 3)),
    ('end_hit', 'f4', (1, 3)),
    ('dir', 'f4', (1, 3)),
    ('enddir', 'f4', (1, 3)),
    ('dir_hit', 'f4', (1, 3)),
    ('enddir_hit', 'f4', (1, 3)),
    ('Evis', 'f4'),
    ('qual', 'f4'),
    ('len_gcm2', 'f4'),
    ('len_cm', 'f4'),
    ('E', 'f4'),
    ('truth', 'i4', (1, 20)),
    ('truthOverlap', 'f4', (1, 20)),
])

particle_data_type = np.dtype([
    ('event_id', 'i4'),
    ('particle_id', 'i4'),
    ('vertex_id', 'i8'),
    ('primary', 'i4'),
    ('pdg', 'i4'),
    ('tgtA', 'i4'),
    ('score', 'f4'),
    ('Evis', 'f4'),
    ('E', 'f4'),
    ('E_method', 'i4'),
    ('p', 'f4', (1, 3)),
    ('dir_hit', 'f4', (1, 3)),
    ('start', 'f4', (1, 3)),
    ('end', 'f4', (1, 3)),
    ('start_hit', 'f4', (1, 3)),
    ('end_hit', 'f4', (1, 3)),
    ('contained', 'i4'),
    ('truth', 'i4', (1, 20)),
    ('truthOverlap', 'f4', (1, 20)),
])

neutrino_data_type = np.dtype([
    ('event_id', 'i4'),
    ('neutrino_id', 'i4'),
    ('vtx', 'f4', (1, 3)),
    ('dir_lngtrk', 'f4', (1, 3)),
    ('dir_heshw', 'f4', (1, 3)),
    ('nuhyp', 'f4', (1, 20)),
    ('Enu', 'f4'),
    ('E_method', 'i4'),
    ('part', 'i4', (1, 20)),
    ('truth', 'i4', (1, 20)),
    ('truthOverlap', 'f4', (1, 20)),
])

interaction_data_type = np.dtype([
    ('event_id', 'i4'),
    ('interaction_id', 'i4'),
    ('vtx', 'f4', (1, 3)),
    ('dir_lngtrk', 'f4', (1, 3)),
    ('dir_heshw', 'f4', (1, 3)),
    ('nuhyp', 'f4', (1, 20)),
    ('Enu', 'f4'),
    ('E_method', 'i4'),
    ('part', 'i4', (1, 20)),
    ('truth', 'i4', (1, 20)),
    ('truthOverlap', 'f4', (1, 20)),
])

neutrino_event_data_type = np.dtype([
    ('event_id', 'i4'),
    ('neutrino_type', 'i4'),
    ('fiducialized', 'i4')
])

interaction_event_data_type = np.dtype([
    ('event_id', 'i4'),
    ('interaction_type', 'i4'),
    ('fiducialized', 'i4')
])

nar_inelastic_data_type = np.dtype([
    ('event_id', 'i4'),
    ('vertex_id', 'i8'),
    ('proton_id', 'i4'),
    ('proton_xyz_start', 'f4', (1, 3)),
    ('nu_vertex', 'f4', (1, 3)),
    ('proton_total_energy', 'f4'),
    ('proton_vis_energy', 'f4'),
    ('proton_length', 'f4'),
    ('proton_pxyz_start', 'f4', (1, 3)),
    ('proton_pxyz_end', 'f4', (1, 3)),
    ('parent_total_energy', 'f4'),
    ('parent_length', 'f4'),
    ('parent_pxyz_start', 'f4', (1, 3)),
    ('parent_pxyz_end', 'f4', (1, 3)),
    ('nu_proton_dt', 'f4'),
    ('nu_proton_distance', 'f4'),
    ('parent_pdg', 'i4'),
    ('grandparent_pdg', 'i4'),
    ('primary_pdg', 'i4'),
    ('primary_length', 'f4'),
    ('neutrino_vertex_fiducialized', 'i4'),
    ('proton_xyz_start_fiducialized', 'i4'),
    ('proton_xyz_end_fiducialized', 'i4'),
    ('truth_segment_overlap', 'f4'),
    ('best_completeness_cluster', 'i4'),
    ('best_completeness', 'f4'),
    ('best_purity', 'f4')
])

undefined_data_type = np.dtype([
    ('event_id', 'i4'),
    ('vertex_id', 'i8'),
    ('traj_id', 'i8'),
    ('parent_id', 'i8'),
    ('pdg_id', 'i8'),
    ('parent_pdg_id', 'i8'),
    ('ancestor_id', 'i8'),
    ('ancestor_pdg_id', 'i8'),
    ('ancestor_level', 'i4'),
    ('start_process', 'i4'),
    ('start_subprocess', 'i4'),
    ('end_process', 'i4'),
    ('end_subprocess', 'i4'),
    ('parent_start_process', 'i4'),
    ('parent_start_subprocess', 'i4'),
    ('parent_end_process', 'i4'),
    ('parent_end_subprocess', 'i4')
])

output_error_data_type = np.dtype([
    ('event_id', 'i4'),
    ('plugin_name', 'S50'),
    ('plugin_error', 'S500'),
    ('plugin_line_number', 'S50'),
    ('plugin_file_name', 'S50'),
])


"""
Utilities for fitting shower data
"""
import numpy as np
from scipy.interpolate import splprep, splev
import warnings
from scipy.integrate import quad, IntegrationWarning
from sklearn.decomposition import PCA

from arrakis_nd.utils.utils import profiler, integrand


@profiler
def fit_shower(
    shower_t0s,
    shower_xyz,
    ancestor_xyz_begin,
    ancestor_pxyz_begin
):
    """
    This function ...
    """
    try:
        """Normalize pxyz_begin"""
        ancestor_pxyz_begin_unit = ancestor_pxyz_begin / np.linalg.norm(ancestor_pxyz_begin)

        """Project charge onto the direction vector"""
        numerator = np.dot(shower_xyz - ancestor_xyz_begin, ancestor_pxyz_begin_unit)
        denominator = np.dot(ancestor_pxyz_begin_unit, ancestor_pxyz_begin_unit)
        t_values = numerator / denominator

        """Find the furthest point index"""
        furthest_point_index = np.argmax(np.abs(t_values))

        """Find the furthest point on the line"""
        furthest_point_on_line = ancestor_xyz_begin + t_values[furthest_point_index] * ancestor_pxyz_begin_unit

        """Calculate the shower length"""
        shower_length = np.sqrt(
            (ancestor_xyz_begin[0] - furthest_point_on_line[0])**2 +
            (ancestor_xyz_begin[1] - furthest_point_on_line[1])**2 +
            (ancestor_xyz_begin[2] - furthest_point_on_line[2])**2
        )

        """ Calculate the vector from p to each point"""
        difference_vectors = shower_xyz - ancestor_xyz_begin

        """ Calculate the cross product of x_unit with each vector"""
        cross_products = np.cross(difference_vectors, ancestor_pxyz_begin_unit)

        """ Calculate the Euclidean norm of each cross product vector"""
        """ This gives the area of the parallelogram formed by the vector and x_unit"""
        distances = np.linalg.norm(cross_products, axis=1)

        """ Divide by the magnitude of x to get the height of the parallelogram, which is the perpbeginicular distance"""
        perpbeginicular_distances = distances / np.linalg.norm(ancestor_pxyz_begin)

        """ Find the maximum perpbeginicular distance"""
        shower_width = np.max(perpbeginicular_distances)
        width_point = np.argmax(perpbeginicular_distances)
        width_point_on_line = ancestor_xyz_begin + t_values[width_point] * ancestor_pxyz_begin_unit
        width_distance = np.sqrt(
            (ancestor_xyz_begin[0] - width_point_on_line[0])**2 +
            (ancestor_xyz_begin[1] - width_point_on_line[1])**2 +
            (ancestor_xyz_begin[2] - width_point_on_line[2])**2
        )
        shower_angle = np.arctan(width_distance / shower_width)
        """Compute the shower angle"""

        """Sort charge with respect to t0s"""
        combined = sorted(zip(shower_t0s, shower_xyz), key=lambda x: x[0])
        sorted_t0, sorted_xyz = zip(*combined)
        sorted_t0 = np.array(sorted_t0)
        sorted_xyz = np.array(sorted_xyz)

        """Fill the values"""
        shower_data = {
            'shower_end': furthest_point_on_line,
            'shower_dir': [0, 0, 0],
            'len_gcm2': 0,
            'len_cm': shower_length,
            'width': shower_width,
            'angle': shower_angle,
        }
    except Exception:
        """Otherwise, these values are undefined"""
        shower_data = {
            'shower_end': [0, 0, 0],
            'shower_dir': [0, 0, 0],
            'len_gcm2': 0,
            'len_cm': 0,
            'width': 0,
            'angle': 0,
        }

    return shower_data



"""
"""
import h5py
import numpy as np

from arrakis_nd.utils.utils import profiler
from arrakis_nd.plugins.plugin import Plugin


class AncestorPlugin(Plugin):
    """
    A plugin for constructing a map of the ancestor information
    """
    def __init__(
        self,
        config: dict = {}
    ):
        """
        """
        super(AncestorPlugin, self).__init__(config)

        self.input_products = None
        self.output_products = [
            'ancestor_traj_index_map',
            'ancestor_traj_id_map',
            'ancestor_pdg_id_map',
            'ancestor_level_map'
        ]

    @profiler
    def process_event(
        self,
        event: int,
        flow_file: h5py.File,
        arrakis_file: h5py.File,
        event_indices: dict,
        event_products: dict,
    ):
        """
        """
        trajectories = flow_file['mc_truth/trajectories/data'][event_indices['trajectories']]

        traj_ids = trajectories['traj_id']
        vertex_ids = trajectories['vertex_id']
        parent_ids = trajectories['parent_id']
        pdg_ids = trajectories['pdg_id']

        """Create empty map with (traj_id, vertex_id) as keys"""
        ancestor_traj_index_map = {
            (traj_id, vertex_id): -1
            for (traj_id, vertex_id) in zip(traj_ids, vertex_ids)
        }
        ancestor_traj_id_map = {
            (traj_id, vertex_id): -1
            for (traj_id, vertex_id) in zip(traj_ids, vertex_ids)
        }
        ancestor_pdg_id_map = {
            (traj_id, vertex_id): -1
            for (traj_id, vertex_id) in zip(traj_ids, vertex_ids)
        }
        ancestor_level_map = {
            (traj_id, vertex_id): 0
            for (traj_id, vertex_id) in zip(traj_ids, vertex_ids)
        }

        for ii, (traj_id, vertex_id, parent_id) in enumerate(zip(
            traj_ids,
            vertex_ids,
            parent_ids
        )):
            """If parent_id == -1, then this is a primary"""
            if parent_id == -1:
                ancestor_traj_index_map[(traj_id, vertex_id)] = ii
                ancestor_traj_id_map[(traj_id, vertex_id)] = traj_id
                ancestor_pdg_id_map[(traj_id, vertex_id)] = pdg_ids[ii]
                ancestor_level_map[(traj_id, vertex_id)] = 0
                continue

            """Otherwise, continue to loop back through until we get to a primary"""
            temp_parent_id = parent_id
            parent_index = -1
            ancestor_level = 0
            while temp_parent_id != -1:
                ancestor_level += 1
                parent_index = np.where(
                    (traj_ids == temp_parent_id) & (vertex_ids == vertex_id)
                )[0][0]
                temp_parent_id = parent_ids[parent_index]
                if temp_parent_id != -1:
                    if ancestor_traj_id_map[(temp_parent_id, vertex_id)] != -1:
                        parent_index = ancestor_traj_index_map[(temp_parent_id, vertex_id)]
                        ancestor_level = ancestor_level_map[(temp_parent_id, vertex_id)] + 1
                        temp_parent_id = -1

            """Assign values from index to maps"""
            ancestor_traj_index_map[(traj_id, vertex_id)] = parent_index
            ancestor_traj_id_map[(traj_id, vertex_id)] = traj_ids[parent_index]
            ancestor_pdg_id_map[(traj_id, vertex_id)] = pdg_ids[parent_index]
            ancestor_level_map[(traj_id, vertex_id)] = ancestor_level

        event_products['ancestor_traj_index_map'] = ancestor_traj_index_map
        event_products['ancestor_traj_id_map'] = ancestor_traj_id_map
        event_products['ancestor_pdg_id_map'] = ancestor_pdg_id_map
        event_products['ancestor_level_map'] = ancestor_level_map



"""
"""
import h5py
import numpy as np

from arrakis_nd.utils.utils import profiler
from arrakis_nd.plugins.plugin import Plugin
from arrakis_nd.utils.track_utils import fit_track
from arrakis_nd.dataset.common import (
    SubProcessType,
    Topology, Physics,
    Fragment
)
from arrakis_nd.arrakis.common import (
    fragment_data_type,
    blip_data_type
)


class PhotonPlugin(Plugin):
    """_summary_

    Args:
        Plugin (_type_): _description_
    """
    def __init__(
        self,
        config: dict = {}
    ):
        """
        """
        super(PhotonPlugin, self).__init__(config)

        self.input_products = [
            'daughters',
            'track_id_hit_map',
            'track_id_hit_segment_map',
            'track_id_hit_t0_map',
            'parent_pdg_id',
        ]
        self.output_products = ['fragment', 'blip']

        if "shower_size_threshold" not in self.config:
            self.config["shower_size_threshold"] = 10
        self.shower_size_threshold = self.config["shower_size_threshold"]

        self.photon_shower_labels = {
            'topology': Topology.Shower.value,
            'physics': Physics.GammaCompton.value,
        }
        self.photon_blip_labels = {
            'topology': Topology.Blip.value,
            'physics': Physics.GammaCompton.value
        }

    @profiler
    def process_event(
        self,
        event: int,
        flow_file: h5py.File,
        arrakis_file: h5py.File,
        event_indices: dict,
        event_products: dict,
    ):
        """
        """
        trajectories = flow_file['mc_truth/trajectories/data'][event_indices['trajectories']]
        charge = flow_file['charge/calib_final_hits/data'][event_indices['charge']]
        arrakis_charge = arrakis_file['charge_segment/calib_final_hits/data'][event_indices['charge']]
        track_id_hit_map = event_products['track_id_hit_map']
        track_id_hit_segment_map = event_products['track_id_hit_segment_map']
        track_id_hit_t0_map = event_products['track_id_hit_t0_map']
        ancestor_traj_id_map = event_products['ancestor_traj_id_map']

        trajectories_traj_ids = trajectories['traj_id']
        trajectories_vertex_ids = trajectories['vertex_id']
        trajectories_pdg_ids = trajectories['pdg_id']
        trajectories_xyz_start = trajectories['xyz_start']
        trajectories_xyz_end = trajectories['xyz_end']
        trajectories_pxyz_start = trajectories['pxyz_start']
        trajectories_pxyz_end = trajectories['pxyz_end']
        trajectories_start_subprocess = trajectories['start_subprocess']
        trajectories_E = trajectories['E_start']
        charge_x = charge['x']
        charge_y = charge['y']
        charge_z = charge['z']
        charge_E = charge['E']
        charge_io_group = charge['io_group']

        """Grab the photons"""
        particle_mask = (
            (abs(trajectories_pdg_ids) == 22)
        )

        """Iterate over the individual (particle_id, vertex_id, traj_index)"""
        for ii, (particle_id, vertex_id, traj_index) in enumerate(zip(
            trajectories_traj_ids[particle_mask],
            trajectories_vertex_ids[particle_mask],
            event_indices['trajectories'][particle_mask]
        )):
            """Get indices in charge_segments arrays"""
            particle_hits = track_id_hit_map[(particle_id, vertex_id)]

            """Check if mip has no hits, go to the next particle"""
            if not any(particle_hits):
                continue

            particle_segments = track_id_hit_segment_map[(particle_id, vertex_id)]
            particle_hit_segments = (
                particle_hits, particle_segments
            )

            """Get the associated t0 values"""
            particle_hit_t0s = track_id_hit_t0_map[(particle_id, vertex_id)]

            """Get ancestor traj_id"""
            ancestor_id = ancestor_traj_id_map[(particle_id, vertex_id)]

            particle_charge_xyz = np.array([
                charge_x[particle_hits],
                charge_y[particle_hits],
                charge_z[particle_hits]
            ]).T
            particle_charge_E = charge_E[particle_hits]
            particle_xyz_start = trajectories_xyz_start[particle_mask][ii]
            particle_xyz_end = trajectories_xyz_end[particle_mask][ii]
            particle_pxyz_start = trajectories_pxyz_start[particle_mask][ii]
            particle_pxyz_end = trajectories_pxyz_end[particle_mask][ii]
            particle_E = trajectories_E[particle_mask][ii]
            """Determine track begin and end points"""
            """
            To do this we find the closest hits to the actual track beginning
            and ending from the trajectories.  If these points end up being the
            same, we skip this track, since there's no way to distinguish
            that point as a track anyways.
            """
            particle_start_distances = np.sqrt(
                ((particle_charge_xyz - particle_xyz_start) ** 2).sum(axis=1)
            )
            particle_end_distances = np.sqrt(
                ((particle_charge_xyz - particle_xyz_end) ** 2).sum(axis=1)
            )

            """############################### Set standard labels ###############################"""
            """Set the event_id"""
            arrakis_charge['event_id'][particle_hits] = event

            """Variable determining if this particle made a fragment or blip"""
            fragment_type = Fragment.Undefined.value
            if (trajectories_start_subprocess[particle_mask][ii] == SubProcessType.Bremsstrahlung.value):
                fragment_type = Fragment.Bremsstrahlung.value
            elif (trajectories_start_subprocess[particle_mask][ii] == SubProcessType.Annihilation.value):
                fragment_type = Fragment.Annihilation.value
            else:
                fragment_type = Fragment.GammaCompton.value
                """This is either neutron capture or hadron inelastic"""

            """############################### Photons ###############################"""
            is_fragment = True
            if len(particle_hits) >= self.shower_size_threshold:
                for label, value in self.photon_shower_labels.items():
                    arrakis_charge[label][particle_hit_segments] = value
            else:
                is_fragment = False
                for label, value in self.photon_blip_labels.items():
                    arrakis_charge[label][particle_hit_segments] = value

            """Set the particle label"""
            arrakis_charge['particle'][particle_hit_segments] = trajectories_pdg_ids[particle_mask][ii]

            """Generate unique label"""
            arrakis_charge['unique_topology'][particle_hit_segments] = traj_index

            """############################### Construct Fragments ###############################"""

            """
            Now we must determine if the fragment is broken along its
            trajectory.  We can do this by checking if the beginning and
            ending points are in different TPCs.  If so, then we will have
            to determine the beginning and ending points of each fragment.

            We can do this by isolating hits according to the io_group,
            which defines the particular TPC region that a hit is in.  For
            each unique io_group associated to the hits of this particle,
            we will follow the same procedure and find the closest
            points to the track beginning and ending to get the fragment
            begin and end points.
            """
            particle_io_group = charge_io_group[particle_hits]
            unique_particle_io_group = np.unique(particle_io_group)
            particle_hits = np.array(particle_hits)
            particle_segments = np.array(particle_segments)
            particle_hit_t0s = np.array(particle_hit_t0s)
            for io_group in unique_particle_io_group:
                """Isolate hits from this io_group"""
                io_group_mask = (particle_io_group == io_group)
                if not sum(io_group_mask):
                    continue
                io_group_hits = particle_hits[io_group_mask]
                io_group_segments = particle_segments[io_group_mask]
                io_group_t0s = particle_hit_t0s[io_group_mask]

                """Find the closest points to track begin/end for this io_group"""
                io_group_start_distances = particle_start_distances[io_group_mask]
                io_group_end_distances = particle_end_distances[io_group_mask]
                fragment_start_index = np.argmin(io_group_start_distances)
                fragment_end_index = np.argmin(io_group_end_distances)

                io_group_start_index = (
                    io_group_hits[fragment_start_index],
                    io_group_segments[fragment_start_index]
                )
                io_group_end_index = (
                    io_group_hits[fragment_end_index],
                    io_group_segments[fragment_end_index]
                )

                if (fragment_start_index != fragment_end_index) and (is_fragment):
                    arrakis_charge['fragment_begin'][io_group_start_index] = 1
                    arrakis_charge['fragment_end'][io_group_end_index] = 1

                """Parameterize the trajectory of this fragment using t0"""
                io_group_xyz = particle_charge_xyz[io_group_mask]
                track_data = fit_track(io_group_t0s, io_group_xyz)

                """Now generate the associated fragment CAF object"""
                """
                The fragment data object has the following entries
                that must be filled (see arrakis.common).
                """
                io_group_xyz = particle_charge_xyz[io_group_mask]

                if is_fragment:
                    """Create the CAF object for this fragment"""
                    fragment_data = np.array([(
                        event,
                        io_group,
                        particle_id,
                        vertex_id,
                        ancestor_id,
                        fragment_type,
                        particle_xyz_start,
                        particle_xyz_end,
                        [io_group_xyz[fragment_start_index]],
                        [io_group_xyz[fragment_end_index]],
                        particle_pxyz_start,
                        particle_pxyz_end,
                        [track_data['track_dir']],
                        [track_data['track_enddir']],
                        sum(particle_charge_E[io_group_mask]),
                        1,
                        track_data['track_len_gcm2'],
                        track_data['track_len_cm'],
                        particle_E,
                        [[traj_index]+[0]*19],
                        [[1]+[0]*19])],
                        dtype=fragment_data_type
                    )

                    """Add the new fragment to the CAF objects"""
                    event_products['fragment'].append(fragment_data)
                else:
                    """Create the CAF object for this fragment"""
                    blip_data = np.array([(
                        event,
                        io_group,
                        particle_id,
                        vertex_id,
                        ancestor_id,
                        fragment_type,
                        particle_xyz_start,
                        particle_xyz_end,
                        [io_group_xyz[fragment_start_index]],
                        [io_group_xyz[fragment_end_index]],
                        particle_pxyz_start,
                        particle_pxyz_end,
                        [track_data['track_dir']],
                        [track_data['track_enddir']],
                        sum(particle_charge_E[io_group_mask]),
                        1,
                        track_data['track_len_gcm2'],
                        track_data['track_len_cm'],
                        particle_E,
                        [[traj_index]+[0]*19],
                        [[1]+[0]*19])],
                        dtype=blip_data_type
                    )

                    """Add the new blip to the CAF objects"""
                    event_products['blip'].append(blip_data)

        """Write changes to arrakis_file"""
        arrakis_file['charge_segment/calib_final_hits/data'][event_indices['charge']] = arrakis_charge



"""
"""
import h5py
import numpy as np

from arrakis_nd.utils.utils import profiler
from arrakis_nd.utils.track_utils import fit_track
from arrakis_nd.plugins.plugin import Plugin
from arrakis_nd.dataset.common import (
    SubProcessType,
    Topology, Physics,
    Fragment
)
from arrakis_nd.arrakis.common import (
    fragment_data_type,
    blip_data_type
)


class FragmentPlugin(Plugin):
    """
    The fragment plugin looks at electrons whose parents are gammas,
    or whose processes originate from

        (1) compton scatters
        (2) gamma conversions
        (3) photoelectric effect
        (4) ionization from another electron

    Each of the electrons/positrons which originate from gamma conversions
    will have a particular size depending on how much initial energy
    they have.  We must decide for each instance of this type of fragment
    whether to consider it as an actual fragment, or as a blip.

    Args:
        Plugin (_type_): _description_
    """
    def __init__(
        self,
        config: dict = {}
    ):
        """
        """
        super(FragmentPlugin, self).__init__(config)

        self.input_products = [
            'daughters',
            'track_id_hit_map',
            'track_id_hit_segment_map',
            'track_id_hit_t0_map',
            'parent_pdg_id',
        ]
        self.output_products = ['fragment', 'blip']

        if "compton_blip_threshold" not in self.config:
            self.config["compton_blip_threshold"] = 10
        self.compton_blip_threshold = self.config["compton_blip_threshold"]

        if "conversion_blip_threshold" not in self.config:
            self.config["conversion_blip_threshold"] = 10
        self.conversion_blip_threshold = self.config["conversion_blip_threshold"]

        if "others_blip_threshold" not in self.config:
            self.config["others_blip_threshold"] = 10
        self.others_blip_threshold = self.config["others_blip_threshold"]

        self.compton_labels = {
            'topology': Topology.Shower.value,
            'physics': Physics.GammaCompton.value,
        }
        self.compton_blip_labels = {
            'topology': Topology.Blip.value,
            'physics': Physics.GammaCompton.value,
        }
        self.conversion_labels = {
            'topology': Topology.Shower.value,
            'physics': Physics.GammaConversion.value,
        }
        self.conversion_blip_labels = {
            'topology': Topology.Blip.value,
            'physics': Physics.GammaConversion.value,
        }
        self.others_fragment_labels = {
            'topology': Topology.Shower.value,
            'physics': Physics.ElectronIonization.value,
        }
        self.others_blip_labels = {
            'topology': Topology.Blip.value,
            'physics': Physics.ElectronIonization.value,
        }

    @profiler
    def process_event(
        self,
        event: int,
        flow_file: h5py.File,
        arrakis_file: h5py.File,
        event_indices: dict,
        event_products: dict,
    ):
        """
        A fragment is part of an electromagnetic cascade which corresponds to an object that
        is contiguous in its energy depositions.  This will be true for electrons/positrons
        which are created in a gamma conversion, but there will also be things which look
        contiguous when you have a series of compton scatters that are close together.
        """
        trajectories = flow_file['mc_truth/trajectories/data'][event_indices['trajectories']]
        charge = flow_file['charge/calib_final_hits/data'][event_indices['charge']]
        arrakis_charge = arrakis_file['charge_segment/calib_final_hits/data'][event_indices['charge']]
        track_id_hit_map = event_products['track_id_hit_map']
        track_id_hit_segment_map = event_products['track_id_hit_segment_map']
        track_id_hit_t0_map = event_products['track_id_hit_t0_map']
        parent_pdg_ids = event_products['parent_pdg_id']
        ancestor_traj_id_map = event_products['ancestor_traj_id_map']

        trajectories_traj_ids = trajectories['traj_id']
        trajectories_vertex_ids = trajectories['vertex_id']
        trajectories_pdg_ids = trajectories['pdg_id']
        trajectories_xyz_start = trajectories['xyz_start']
        trajectories_xyz_end = trajectories['xyz_end']
        trajectories_pxyz_start = trajectories['pxyz_start']
        trajectories_pxyz_end = trajectories['pxyz_end']
        trajectories_start_subprocess = trajectories['start_subprocess']
        trajectories_E = trajectories['E_start']
        charge_x = charge['x']
        charge_y = charge['y']
        charge_z = charge['z']
        charge_E = charge['E']
        charge_io_group = charge['io_group']

        """Grab the electrons which are from compton scatters, conversions or electron ionization"""
        """
        We are looking for electrons which also:
            (1) have a gamma parent, or
            (2) come from a compton scatter, or
            (3) come from a gamma conversion, or
            (4) come from a pair production by charge, or
            (5) have an electron parent and
                (5a) come from the photo electric effect, or
                (5b) come from ionization,
            (6) come from a pi0 decay

        The reason we must consider the different subprocesses, event though we are already
        looking for gamma parents, is that edep-sim does some bookkeeping that may remove
        information about intermediate gammas which leave no energy deposits.  Thus, an
        electron can have a gamma conversion as its subprocess, but will be linked to the parent
        of the gamma (which may be a mip or hip, etc.).

        The same thing can happen when the parent is an electron which bremsstrahlungs and
        then the corresponding gamma does something like the photoelectric effect.
        """
        particle_mask = (
            (abs(trajectories_pdg_ids) == 11) &
            (
                (abs(parent_pdg_ids) == 22) |
                (abs(trajectories_start_subprocess) == SubProcessType.ComptonScattering.value) |
                (abs(trajectories_start_subprocess) == SubProcessType.GammaConversion.value) |
                (abs(trajectories_start_subprocess) == SubProcessType.PairProdByCharge.value) |
                (
                    (abs(parent_pdg_ids) == 11) &
                    (
                        (abs(trajectories_start_subprocess) == SubProcessType.PhotoElectricEffect.value) |
                        (abs(trajectories_start_subprocess) == SubProcessType.Ionization.value)
                    )
                ) |
                (
                    (abs(parent_pdg_ids) == 111)
                )
            )
        )

        """Iterate over the individual (particle_id, vertex_id, traj_index)"""
        for ii, (particle_id, vertex_id, traj_index) in enumerate(zip(
            trajectories_traj_ids[particle_mask],
            trajectories_vertex_ids[particle_mask],
            event_indices['trajectories'][particle_mask]
        )):
            """Get indices in charge_segments arrays"""
            particle_hits = track_id_hit_map[(particle_id, vertex_id)]

            """Check if mip has no hits, go to the next particle"""
            if not any(particle_hits):
                continue

            particle_segments = track_id_hit_segment_map[(particle_id, vertex_id)]
            particle_hit_segments = (
                particle_hits, particle_segments
            )

            """Get the associated t0 values"""
            particle_hit_t0s = track_id_hit_t0_map[(particle_id, vertex_id)]

            """Get ancestor traj_id"""
            ancestor_id = ancestor_traj_id_map[(particle_id, vertex_id)]

            """############################### Set standard labels ###############################"""
            """Set the event_id"""
            arrakis_charge['event_id'][particle_hits] = event

            """############################### Electrons ###############################"""
            """

            """
            particle_charge_xyz = np.array([
                charge_x[particle_hits],
                charge_y[particle_hits],
                charge_z[particle_hits]
            ]).T
            particle_charge_E = charge_E[particle_hits]
            particle_xyz_start = trajectories_xyz_start[particle_mask][ii]
            particle_xyz_end = trajectories_xyz_end[particle_mask][ii]
            particle_pxyz_start = trajectories_pxyz_start[particle_mask][ii]
            particle_pxyz_end = trajectories_pxyz_end[particle_mask][ii]
            particle_E = trajectories_E[particle_mask][ii]
            """Determine track begin and end points"""
            """
            To do this we find the closest hits to the actual track beginning
            and ending from the trajectories.  If these points end up being the
            same, we skip this track, since there's no way to distinguish
            that point as a track anyways.
            """
            particle_start_distances = np.sqrt(
                ((particle_charge_xyz - particle_xyz_start) ** 2).sum(axis=1)
            )
            particle_end_distances = np.sqrt(
                ((particle_charge_xyz - particle_xyz_end) ** 2).sum(axis=1)
            )
            closest_start_index = np.argmin(particle_start_distances)
            closest_end_index = np.argmin(particle_end_distances)

            particle_start_index = (
                particle_hits[closest_start_index],
                particle_segments[closest_start_index]
            )
            particle_end_index = (
                particle_hits[closest_end_index],
                particle_segments[closest_end_index]
            )

            """Variable determining if this particle made a fragment or blip"""
            is_fragment = True
            fragment_type = Fragment.Undefined.value

            """Assign labels based on particle type"""
            if (
                (abs(trajectories_start_subprocess[particle_mask][ii]) == SubProcessType.ComptonScattering.value)
            ):
                fragment_type = Fragment.GammaCompton.value
                """Iterate over compton scattering labels"""
                if len(particle_hits) < self.compton_blip_threshold:
                    is_fragment = False
                    for label, value in self.compton_blip_labels.items():
                        arrakis_charge[label][particle_hit_segments] = value
                else:
                    if closest_start_index != closest_end_index:
                        arrakis_charge['fragment_begin'][particle_start_index] = 1
                        arrakis_charge['fragment_end'][particle_end_index] = 1
                    for label, value in self.compton_labels.items():
                        arrakis_charge[label][particle_hit_segments] = value
            elif (
                (abs(trajectories_start_subprocess[particle_mask][ii]) == SubProcessType.GammaConversion.value) |
                (abs(trajectories_start_subprocess[particle_mask][ii]) == SubProcessType.PairProdByCharge.value) |
                (
                    (abs(parent_pdg_ids[particle_mask][ii]) == 111) &
                    (abs(trajectories_start_subprocess[particle_mask][ii] == SubProcessType.Decay.value))
                )
            ):
                if (abs(trajectories_start_subprocess[particle_mask][ii]) == SubProcessType.GammaConversion.value):
                    fragment_type = Fragment.GammaConversion.value
                elif (abs(trajectories_start_subprocess[particle_mask][ii]) == SubProcessType.PairProdByCharge.value):
                    fragment_type = Fragment.PairProductionByCharge.value
                else:
                    fragment_type = Fragment.GammaConversion.value
                """If this particle is a conversion electron, add labels and CAF"""
                if len(particle_hits) < self.conversion_blip_threshold:
                    is_fragment = False
                    for label, value in self.conversion_blip_labels.items():
                        arrakis_charge[label][particle_hit_segments] = value
                else:
                    if closest_start_index != closest_end_index:
                        arrakis_charge['fragment_begin'][particle_start_index] = 1
                        arrakis_charge['fragment_end'][particle_end_index] = 1
                    for label, value in self.conversion_labels.items():
                        arrakis_charge[label][particle_hit_segments] = value
            else:
                """
                Otherwise, consider this particle a low energy scatter.
                """
                """Iterate over standard labels"""
                if (abs(trajectories_start_subprocess[particle_mask][ii]) == SubProcessType.PhotoElectricEffect.value):
                    fragment_type = Fragment.PhotoelectricEffect.value
                else:
                    fragment_type = Fragment.ElectronIonization.value
                if len(particle_hits) < self.others_blip_threshold:
                    is_fragment = False
                    for label, value in self.others_blip_labels.items():
                        arrakis_charge[label][particle_hit_segments] = value
                else:
                    if closest_start_index != closest_end_index:
                        arrakis_charge['fragment_begin'][particle_start_index] = 1
                        arrakis_charge['fragment_end'][particle_end_index] = 1
                    for label, value in self.others_fragment_labels.items():
                        arrakis_charge[label][particle_hit_segments] = value

            """Set the particle label"""
            arrakis_charge['particle'][particle_hit_segments] = trajectories_pdg_ids[particle_mask][ii]

            """Generate unique label"""
            arrakis_charge['unique_topology'][particle_hit_segments] = traj_index

            """############################### Construct Fragments ###############################"""

            """
            Now we must determine if the fragment is broken along its
            trajectory.  We can do this by checking if the beginning and
            ending points are in different TPCs.  If so, then we will have
            to determine the beginning and ending points of each fragment.

            We can do this by isolating hits according to the io_group,
            which defines the particular TPC region that a hit is in.  For
            each unique io_group associated to the hits of this particle,
            we will follow the same procedure and find the closest
            points to the track beginning and ending to get the fragment
            begin and end points.
            """
            particle_io_group = charge_io_group[particle_hits]
            unique_particle_io_group = np.unique(particle_io_group)
            particle_hits = np.array(particle_hits)
            particle_segments = np.array(particle_segments)
            particle_hit_t0s = np.array(particle_hit_t0s)
            particle_fragment_ids = []
            for io_group in unique_particle_io_group:
                """Isolate hits from this io_group"""
                io_group_mask = (particle_io_group == io_group)
                if not sum(io_group_mask):
                    continue
                io_group_hits = particle_hits[io_group_mask]
                io_group_segments = particle_segments[io_group_mask]
                io_group_t0s = particle_hit_t0s[io_group_mask]

                """Find the closest points to track begin/end for this io_group"""
                io_group_start_distances = particle_start_distances[io_group_mask]
                io_group_end_distances = particle_end_distances[io_group_mask]
                fragment_start_index = np.argmin(io_group_start_distances)
                fragment_end_index = np.argmin(io_group_end_distances)

                io_group_start_index = (
                    io_group_hits[fragment_start_index],
                    io_group_segments[fragment_start_index]
                )
                io_group_end_index = (
                    io_group_hits[fragment_end_index],
                    io_group_segments[fragment_end_index]
                )

                if (fragment_start_index != fragment_end_index) and (is_fragment):
                    arrakis_charge['fragment_begin'][io_group_start_index] = 1
                    arrakis_charge['fragment_end'][io_group_end_index] = 1

                """Parameterize the trajectory of this fragment using t0"""
                io_group_xyz = particle_charge_xyz[io_group_mask]
                track_data = fit_track(io_group_t0s, io_group_xyz)

                """Now generate the associated fragment CAF object"""
                """
                The fragment data object has the following entries
                that must be filled (see arrakis.common).
                """
                io_group_xyz = particle_charge_xyz[io_group_mask]

                """Assign this fragment to the mip track"""
                particle_fragment_ids.append(io_group_end_index[0])

                if is_fragment:
                    """Create the CAF object for this fragment"""
                    fragment_data = np.array([(
                        event,
                        io_group,
                        particle_id,
                        vertex_id,
                        ancestor_id,
                        fragment_type,
                        particle_xyz_start,
                        particle_xyz_end,
                        [io_group_xyz[fragment_start_index]],
                        [io_group_xyz[fragment_end_index]],
                        particle_pxyz_start,
                        particle_pxyz_end,
                        [track_data['track_dir']],
                        [track_data['track_enddir']],
                        sum(particle_charge_E[io_group_mask]),
                        1,
                        track_data['track_len_gcm2'],
                        track_data['track_len_cm'],
                        particle_E,
                        [[traj_index]+[0]*19],
                        [[1]+[0]*19])],
                        dtype=fragment_data_type
                    )

                    """Add the new fragment to the CAF objects"""
                    event_products['fragment'].append(fragment_data)
                else:
                    """Create the CAF object for this fragment"""
                    blip_data = np.array([(
                        event,
                        io_group,
                        particle_id,
                        vertex_id,
                        ancestor_id,
                        fragment_type,
                        particle_xyz_start,
                        particle_xyz_end,
                        [io_group_xyz[fragment_start_index]],
                        [io_group_xyz[fragment_end_index]],
                        particle_pxyz_start,
                        particle_pxyz_end,
                        [track_data['track_dir']],
                        [track_data['track_enddir']],
                        sum(particle_charge_E[io_group_mask]),
                        1,
                        track_data['track_len_gcm2'],
                        track_data['track_len_cm'],
                        particle_E,
                        [[traj_index]+[0]*19],
                        [[1]+[0]*19])],
                        dtype=blip_data_type
                    )

                    """Add the new blip to the CAF objects"""
                    event_products['blip'].append(blip_data)

        """Write changes to arrakis_file"""
        arrakis_file['charge_segment/calib_final_hits/data'][event_indices['charge']] = arrakis_charge



"""
"""
import h5py
import numpy as np

from arrakis_nd.utils.utils import profiler
from arrakis_nd.utils.track_utils import fit_track
from arrakis_nd.plugins.plugin import Plugin
from arrakis_nd.dataset.common import (
    Topology, Physics,
    Track
)
from arrakis_nd.arrakis.common import (
    tracklette_data_type,
    track_data_type
)


class TrackPlugin(Plugin):
    """
    A plugin for labeling mips in charge and light data.

    mips are MIPs which deposit energy through MIP ionization (i.e. the
    standard dE/dx mechanism).  This plugin grabs all mips in an event
    and labels their energy deposits according to the following scheme:

        (1) hits coming directly from mip ionization are labeled as
                topology = track
                physics = mip
            mips are defined as muons, tauons, pions and kaons
        (2) hits coming directly from hip ionization are labeled as
                topology = track
                physics = hip
            hips are defined as protons

    Apart from these labels, track beginning and ending points are also
    labelled.

    StandardRecord objects are also generated.
    """
    def __init__(
        self,
        config: dict = {}
    ):
        """
        """
        super(TrackPlugin, self).__init__(config)

        self.input_products = [
            'daughters',
            'track_id_hit_map',
            'track_id_hit_segment_map',
            'track_id_hit_t0_map'
        ]
        self.output_products = ['tracklette', 'track']

        self.mip_labels = {
            'topology': Topology.Track.value,
            'physics': Physics.MIP.value,
        }
        self.hip_labels = {
            'topology': Topology.Track.value,
            'physics': Physics.HIP.value,
        }

    @profiler
    def process_event(
        self,
        event: int,
        flow_file: h5py.File,
        arrakis_file: h5py.File,
        event_indices: dict,
        event_products: dict,
    ):
        """
        """
        trajectories = flow_file['mc_truth/trajectories/data'][event_indices['trajectories']]
        charge = flow_file['charge/calib_final_hits/data'][event_indices['charge']]
        arrakis_charge = arrakis_file['charge_segment/calib_final_hits/data'][event_indices['charge']]
        track_id_hit_map = event_products['track_id_hit_map']
        track_id_hit_segment_map = event_products['track_id_hit_segment_map']
        track_id_hit_t0_map = event_products['track_id_hit_t0_map']

        trajectories_traj_ids = trajectories['traj_id']
        trajectories_vertex_ids = trajectories['vertex_id']
        trajectories_pdg_ids = trajectories['pdg_id']
        trajectories_xyz_start = trajectories['xyz_start']
        trajectories_xyz_end = trajectories['xyz_end']
        trajectories_pxyz_start = trajectories['pxyz_start']
        trajectories_pxyz_end = trajectories['pxyz_end']
        trajectories_E = trajectories['E_start']
        charge_x = charge['x']
        charge_y = charge['y']
        charge_z = charge['z']
        charge_E = charge['E']
        charge_io_group = charge['io_group']

        """Grab the mips/hips"""
        particle_mask = (
            (abs(trajectories_pdg_ids) == 13) |
            (abs(trajectories_pdg_ids) == 15) |
            (abs(trajectories_pdg_ids) == 211) |
            (abs(trajectories_pdg_ids) == 321) |
            (abs(trajectories_pdg_ids) == 2212)
        )

        """Iterate over the individual (particle_id, vertex_id, traj_index)"""
        for ii, (particle_id, vertex_id, traj_index) in enumerate(zip(
            trajectories_traj_ids[particle_mask],
            trajectories_vertex_ids[particle_mask],
            event_indices['trajectories'][particle_mask]
        )):
            """Get indices in charge_segments arrays"""
            particle_hits = track_id_hit_map[(particle_id, vertex_id)]

            """Check if mip has no hits, go to the next particle"""
            if not any(particle_hits):
                continue

            particle_segments = track_id_hit_segment_map[(particle_id, vertex_id)]
            particle_hit_segments = (
                particle_hits, particle_segments
            )

            """Get the associated t0 values"""
            particle_hit_t0s = track_id_hit_t0_map[(particle_id, vertex_id)]

            """############################### Set standard labels ###############################"""
            """Set the event_id"""
            arrakis_charge['event_id'][particle_hits] = event

            """Iterate over standard labels"""
            if abs(trajectories_pdg_ids[particle_mask][ii]) == 2212:
                track_type = Track.HIP.value
                for label, value in self.hip_labels.items():
                    arrakis_charge[label][particle_hit_segments] = value
            else:
                track_type = Track.MIP.value
                for label, value in self.mip_labels.items():
                    arrakis_charge[label][particle_hit_segments] = value

            """Set the particle label"""
            arrakis_charge['particle'][particle_hit_segments] = trajectories_pdg_ids[particle_mask][ii]

            """Generate unique label"""
            arrakis_charge['unique_topology'][particle_hit_segments] = traj_index

            """Determine track begin and end points"""
            particle_charge_xyz = np.array([
                charge_x[particle_hits],
                charge_y[particle_hits],
                charge_z[particle_hits]
            ]).T
            particle_charge_E = charge_E[particle_hits]
            particle_xyz_start = trajectories_xyz_start[particle_mask][ii]
            particle_xyz_end = trajectories_xyz_end[particle_mask][ii]
            particle_pxyz_start = trajectories_pxyz_start[particle_mask][ii]
            particle_pxyz_end = trajectories_pxyz_end[particle_mask][ii]
            particle_E = trajectories_E[particle_mask][ii]

            """
            To do this we find the closest hits to the actual track beginning
            and ending from the trajectories.  If these points end up being the
            same, we skip this track, since there's no way to distinguish
            that point as a track anyways.
            """
            particle_start_distances = np.sqrt(
                ((particle_charge_xyz - particle_xyz_start) ** 2).sum(axis=1)
            )
            particle_end_distances = np.sqrt(
                ((particle_charge_xyz - particle_xyz_end) ** 2).sum(axis=1)
            )
            closest_start_index = np.argmin(particle_start_distances)
            closest_end_index = np.argmin(particle_end_distances)

            particle_start_index = (
                particle_hits[closest_start_index],
                particle_segments[closest_start_index]
            )
            particle_end_index = (
                particle_hits[closest_end_index],
                particle_segments[closest_end_index]
            )

            if closest_start_index == closest_end_index:
                continue

            """Set track beginning and ending points"""
            arrakis_charge['tracklette_begin'][particle_start_index] = 1
            arrakis_charge['tracklette_end'][particle_end_index] = 1

            """Parameterize the trajectory of this track using t0"""
            track_data = fit_track(particle_hit_t0s, particle_charge_xyz)

            """############################### Construct Tracklettes ###############################"""

            """
            Now we must determine if the track is broken along its
            trajectory.  We can do this by checking if the beginning and
            ending points are in different TPCs.  If so, then we will have
            to determine the beginning and ending points of each tracklette.

            We can do this by isolating hits according to the io_group,
            which defines the particular TPC region that a hit is in.  For
            each unique io_group associated to the hits of this particle,
            we will follow the same procedure and find the closest
            points to the track beginning and ending to get the tracklette
            begin and end points.
            """
            particle_io_group = charge_io_group[particle_hits]
            unique_particle_io_group = np.unique(particle_io_group)
            particle_hits = np.array(particle_hits)
            particle_segments = np.array(particle_segments)
            particle_hit_t0s = np.array(particle_hit_t0s)
            for io_group in unique_particle_io_group:
                """Isolate hits from this io_group"""
                io_group_mask = (particle_io_group == io_group)
                if not sum(io_group_mask):
                    continue
                io_group_hits = particle_hits[io_group_mask]
                io_group_segments = particle_segments[io_group_mask]
                io_group_t0s = particle_hit_t0s[io_group_mask]

                """Find the closest points to track begin/end for this io_group"""
                io_group_start_distances = particle_start_distances[io_group_mask]
                io_group_end_distances = particle_end_distances[io_group_mask]
                tracklette_start_index = np.argmin(io_group_start_distances)
                tracklette_end_index = np.argmin(io_group_end_distances)

                io_group_start_index = (
                    io_group_hits[tracklette_start_index],
                    io_group_segments[tracklette_start_index]
                )
                io_group_end_index = (
                    io_group_hits[tracklette_end_index],
                    io_group_segments[tracklette_end_index]
                )

                if tracklette_start_index == tracklette_end_index:
                    continue

                arrakis_charge['tracklette_begin'][io_group_start_index] = 1
                arrakis_charge['tracklette_end'][io_group_end_index] = 1

                """Parameterize the trajectory of this tracklette using t0"""
                io_group_xyz = particle_charge_xyz[io_group_mask]
                track_data = fit_track(io_group_t0s, io_group_xyz)

                """Now generate the associated tracklette CAF object"""
                """
                The tracklette data object has the following entries
                that must be filled (see arrakis.common).
                """
                io_group_xyz = particle_charge_xyz[io_group_mask]

                """Create the CAF object for this tracklette"""
                tracklette_data = np.array([(
                    event,
                    io_group,
                    particle_id,
                    vertex_id,
                    track_type,
                    [0, 0, 0],
                    [0, 0, 0],
                    [io_group_xyz[tracklette_start_index]],
                    [io_group_xyz[tracklette_end_index]],
                    [0, 0, 0],
                    [0, 0, 0],
                    [track_data['track_dir']],
                    [track_data['track_enddir']],
                    sum(particle_charge_E[io_group_mask]),
                    1,
                    track_data['track_len_gcm2'],
                    track_data['track_len_cm'],
                    0,
                    [[traj_index]+[0]*19],
                    [[1]+[0]*19])],
                    dtype=tracklette_data_type
                )

                """Add the new tracklette to the CAF objects"""
                event_products['tracklette'].append(tracklette_data)

            """############################### Construct Track ###############################"""
            """Now generate the associated track CAF object"""
            """
            The track data object has the following entries
            that must be filled (see arrakis.common).
            """
            io_group_xyz = particle_charge_xyz[io_group_mask]

            """Create the CAF object for this track"""
            particle_pxyz_start_magnitude = np.linalg.norm(particle_pxyz_start)
            particle_pxyz_end_magnitude = np.linalg.norm(particle_pxyz_end)
            if particle_pxyz_start_magnitude > 0:
                particle_pxyz_start /= particle_pxyz_start_magnitude
            if particle_pxyz_end_magnitude > 0:
                particle_pxyz_end /= particle_pxyz_end_magnitude

            track_data = np.array([(
                event,
                particle_id,
                vertex_id,
                track_type,
                particle_xyz_start,
                particle_xyz_end,
                [particle_charge_xyz[closest_start_index]],
                [particle_charge_xyz[closest_end_index]],
                particle_pxyz_start,
                particle_pxyz_end,
                [track_data['track_dir']],
                [track_data['track_enddir']],
                sum(particle_charge_E),
                1,
                track_data['track_len_gcm2'],
                track_data['track_len_cm'],
                particle_E,
                [[traj_index]+[0]*19],
                [[1]+[0]*19])],
                dtype=track_data_type
            )

            """Add the new track to the CAF objects"""
            event_products['track'].append(track_data)

        """Write changes to arrakis_file"""
        arrakis_file['charge_segment/calib_final_hits/data'][event_indices['charge']] = arrakis_charge



    """
    """
    import numpy as np
    
    
    class Geometry:
        """
        This class contains several useful functions related
        to the detector geometry of the various ND TPCs. 
        """
        def __init__(
            self,
            config: dict = {}
        ):
            self.config = config

            """Configuration for 2x2 geometry"""
            self.active_tpc_widths_2x2 = [30.6, 130., 64.]
            self.tpcs_relative_to_module_2x2=[
                [-15.7, 0., 0.],
                [15.7,  0., 0.]
            ]
            self.modules_relative_to_2x2=[
                [-33.5,0., -33.5],
                [33.5,0.,  -33.5],
                [-33.5,0., 33.5],
                [33.5,0.,  33.5]
            ]
            self.detector_center_2x2=[0., -268, 1300]
            
            """Set up detector geometry"""
            self.parse_config()

        def parse_config(self):
            """
            Here we determine what the various detector variables should be,
            and then construct other variables relative to them.
            """
            self.active_tpc_widths = self.active_tpc_widths_2x2
            self.tpcs_relative_to_module = self.tpcs_relative_to_module_2x2
            self.modules_relative_to = self.modules_relative_to_2x2
            self.detector_center = self.detector_center_2x2
        
        def determine_tpc_bounds(i):
            active_tpc_widths=[30.6, 130., 64.] # cm
            tpcs_relative_to_module=[[-15.7,0.,0.],[15.7,0.,0.]]
            modules_relative_to_2x2=[[-33.5,0.,-33.5],
                                    [33.5,0.,-33.5],
                                    [-33.5,0.,33.5],
                                    [33.5,0.,33.5]]
            detector_center=[0.,-268,1300]
            tpc_bounds=np.array([-active_tpc_widths[i]/2., active_tpc_widths[i]/2.])
            tpc_bounds_relative_to_2x2=[]
            for tpc in tpcs_relative_to_module:
                tpc_bound_relative_to_module = tpc_bounds + tpc[i]
                for module in modules_relative_to_2x2:
                    bound = tpc_bound_relative_to_module + module[i]
                    tpc_bounds_relative_to_2x2.append(bound)
            bounds_relative_to_NDhall = np.array(tpc_bounds_relative_to_2x2) + detector_center[i]
            return np.unique(bounds_relative_to_NDhall, axis=0)


"""
"""
import h5py
import numpy as np
from scipy.interpolate import splprep, splev

from arrakis_nd.utils.utils import profiler
from arrakis_nd.utils.utils import fiducialized_vertex
from arrakis_nd.utils.track_utils import fit_track
from arrakis_nd.plugins.plugin import Plugin
from arrakis_nd.arrakis.common import particle_data_type


class ParticlePlugin(Plugin):
    """
    A plugin for creating particle standard record objects.

    """
    def __init__(
        self,
        config: dict = {}
    ):
        """
        A particle standard record object is essentially a fully reconstructed
        particle that can be associated back to the particles in the trajectories
        within mc_truth.  Not all information is (or can be) reconstructed, such
        as the process, endprocess, etc., but basic information about the start
        and end points, the starting energy and momentum and whether the particle
        is contained within the TPC are present.

        Constructing these objects is fairly simple, we just need to loop over
        all particles in the trajectories for this event, and see if each particle
        has reconstructed hits or not. If so, then construct the associated
        particle standard record object.

        Since it is in principle impossible to determine the "actual" start and
        end points of particle trajectories, we instead identify the hits
        which are closest to the start and end points.  Even if one could pin point
        the exact starting location for trajectories which begin inside the detector,
        there is no way to determine this for particles outside.  Likewise for the momentum,
        since particles which begin outside cannot have their momentum reconstructed
        faithfully, we instead determine the momentum from fitting a curve to the
        trajectory of the particle and estimating the gradient at the start point.

        Another quantity which seems rather useless is 'tgtA', which is the target
        nucleus that the particle originated from.  This is also impossible to determine
        outside of the detector, so only argon is a reasonable quantity to use, hence
        tgtA is redundant.
        """
        super(ParticlePlugin, self).__init__(config)

        self.input_products = [
            'daughters',
            'track_id_hit_map',
            'track_id_hit_segment_map',
            'track_id_hit_t0_map'
        ]
        self.output_products = ['particle']

    @profiler
    def process_event(
        self,
        event: int,
        flow_file: h5py.File,
        arrakis_file: h5py.File,
        event_indices: dict,
        event_products: dict,
    ):
        """
        """
        interactions = flow_file['mc_truth/interactions/data'][event_indices['interactions']]
        trajectories = flow_file['mc_truth/trajectories/data'][event_indices['trajectories']]
        charge = flow_file['charge/calib_final_hits/data'][event_indices['charge']]
        track_id_hit_map = event_products['track_id_hit_map']
        track_id_hit_t0_map = event_products['track_id_hit_t0_map']

        trajectories_traj_ids = trajectories['traj_id']
        trajectories_vertex_ids = trajectories['vertex_id']
        trajectories_pdg_ids = trajectories['pdg_id']
        trajectories_parent_ids = trajectories['parent_id']
        trajectories_xyz_start = trajectories['xyz_start']
        trajectories_xyz_end = trajectories['xyz_end']
        trajectories_pxyz_start = trajectories['pxyz_start']
        trajectories_E = trajectories['E_start']
        charge_x = charge['x']
        charge_y = charge['y']
        charge_z = charge['z']
        charge_E = charge['E']

        """Iterate over the individual (traj_id, vertex_id, traj_index)"""
        for ii, (particle_id, vertex_id, traj_index) in enumerate(zip(
            trajectories_traj_ids,
            trajectories_vertex_ids,
            event_indices['trajectories']
        )):
            """Get indices in charge_segments arrays"""
            particle_hits = track_id_hit_map[(particle_id, vertex_id)]

            """Check if particle has no hits, go to the next particle"""
            if not any(particle_hits):
                continue

            """Get the associated t0 values"""
            particle_hit_t0s = track_id_hit_t0_map[(particle_id, vertex_id)]

            """Determine if the particle is a primary"""
            if trajectories_parent_ids[ii] == -1:
                particle_primary = 1
            else:
                particle_primary = 0

            """Find the closest start and end hits"""
            particle_charge_xyz = np.array([
                charge_x[particle_hits],
                charge_y[particle_hits],
                charge_z[particle_hits]
            ]).T
            particle_charge_E = charge_E[particle_hits]
            particle_xyz_start = trajectories_xyz_start[ii]
            particle_xyz_end = trajectories_xyz_end[ii]
            particle_pxyz_start = trajectories_pxyz_start[ii]
            particle_E = trajectories_E[ii]

            """
            To do this we find the closest hits to the actual particle beginning
            and ending from the trajectories.  If these points end up being the
            same, we skip this particle, since there's no way to distinguish
            that point as a particle anyways.
            """
            particle_start_distances = np.sqrt(
                ((particle_charge_xyz - particle_xyz_start) ** 2).sum(axis=1)
            )
            particle_end_distances = np.sqrt(
                ((particle_charge_xyz - particle_xyz_end) ** 2).sum(axis=1)
            )
            closest_start_index = np.argmin(particle_start_distances)
            closest_end_index = np.argmin(particle_end_distances)

            closest_particle_xyz_start = particle_charge_xyz[closest_start_index]
            closest_particle_xyz_end = particle_charge_xyz[closest_end_index]

            """Parameterize the trajectory of this particle using t0"""
            track_data = fit_track(particle_hit_t0s, particle_charge_xyz)

            """Determine tgtA"""
            interaction = interactions[(interactions['vertex_id'] == vertex_id)]
            if fiducialized_vertex(particle_xyz_start):
                tgtA = 18
            else:
                if particle_primary:
                    tgtA = interaction['target']
                else:
                    tgtA = 0

            """Let's create the particle object"""
            """
            Particles in the standard record have the following variables:
                particle_data_type = np.dtype([
                    ('event_id', 'i4'),
                    ('particle_id', 'i4'),
                    ('vertex_id', 'i8'),
                    ('primary', 'i4'),
                    ('pdg', 'i4'),
                    ('tgtA', 'i4'),
                    ('score', 'f4'),
                    ('Evis', 'f4'),
                    ('E', 'f4'),
                    ('E_method', 'i4'),
                    ('p', 'f4', (1, 3)),
                    ('dir_hit', 'f4', (1, 3)),
                    ('start', 'f4', (1, 3)),
                    ('end', 'f4', (1, 3)),
                    ('start_hit', 'f4', (1, 3)),
                    ('end_hit', 'f4', (1, 3)),
                    ('contained', 'i4'),
                    ('truth', 'i4', (1, 20)),
                    ('truthOverlap', 'f4', (1, 20)),
                ])
            """

            particle = np.array([(
                event,
                particle_id,
                vertex_id,
                particle_primary,
                trajectories_pdg_ids[ii],
                tgtA,
                1,
                sum(particle_charge_E),
                particle_E,
                0,
                particle_pxyz_start,
                track_data['track_dir'],
                particle_xyz_start,
                particle_xyz_end,
                closest_particle_xyz_start,
                closest_particle_xyz_end,
                fiducialized_vertex(particle_xyz_start) and fiducialized_vertex(particle_xyz_end),
                [[traj_index]+[0]*19],
                [[1]+[0]*19])],
                dtype=particle_data_type
            )

            """Add new particle to the event products"""
            event_products['particle'].append(particle)



"""
"""
import h5py
import numpy as np

from arrakis_nd.utils.utils import profiler
from arrakis_nd.plugins.plugin import Plugin
from arrakis_nd.dataset.common import (
    Topology, Physics
)
from arrakis_nd.utils.logger import ArrakisError
from arrakis_nd.arrakis.common import undefined_data_type

class ConstraintPlugin(Plugin):
    """
    This plugin takes all the semantic information determine from the other
    plugins and decides how to "officially" label each calib_final_hit.  This is
    done in two steps,

        (1) apply an influence cut based on the distance of each contributing
            segment to the hit.
        (2) make heirarchical choices to the labels based on the following criteria:
            (a) michels take 1st precendence
            (b) deltas take 2nd precendence
            (c) tracks take 3rd precendence
            (d) showers take 4th precendence
            (e) blips take last precendence

    Args:
        Plugin (_type_): _description_
    """
    def __init__(
        self,
        config: dict = {}
    ):
        """
        """
        super(ConstraintPlugin, self).__init__(config)

        self.input_products = [
            'daughters',
            'track_id_hit_map',
            'track_id_hit_segment_map',
            'track_id_hit_t0_map',
            'parent_pdg_id',
        ]
        self.output_products = [
            'undefined'
        ]

        if "segment_influence_cut" not in self.config:
            self.config["segment_influence_cut"] = 1.0
        self.segment_influence_cut = self.config["segment_influence_cut"]

        if "constraint_mode" not in self.config:
            self.config["constraint_mode"] = "min_distance"
        if self.config["constraint_mode"] not in ["max_fraction", "min_distance"]:
            raise ArrakisError(f"specified 'constraint_mode' {self.config['constraint_mode']} not allowed!")
        self.constraint_mode = self.config["constraint_mode"]
        
        if "replace_undefined" not in self.config:
            self.config["replace_undefined"] = -1
        self.replace_undefined = self.config["replace_undefined"]

    @profiler
    def process_event(
        self,
        event: int,
        flow_file: h5py.File,
        arrakis_file: h5py.File,
        event_indices: dict,
        event_products: dict,
    ):
        """
        """
        arrakis_charge = arrakis_file['charge/calib_final_hits/data'][event_indices['charge']]
        arrakis_segment_charge = arrakis_file['charge_segment/calib_final_hits/data'][event_indices['charge']]
        charge_back_track = flow_file['mc_truth/calib_final_hit_backtrack/data'][event_indices['charge']]
        trajectories = flow_file['mc_truth/trajectories/data'][event_indices['trajectories']]
        segments = flow_file['mc_truth/segments/data'][event_indices['segments']]

        trajectories_traj_ids = trajectories['traj_id']
        trajectories_vertex_ids = trajectories['vertex_id']
        trajectories_parent_ids = trajectories['parent_id']
        trajectories_pdg_ids = trajectories['pdg_id']
        trajectories_start_process = trajectories['start_process']
        trajectories_start_subprocess = trajectories['start_subprocess']
        trajectories_end_process = trajectories['end_process']
        trajectories_end_subprocess = trajectories['end_subprocess']
        trajectories_parent_pdg_ids = event_products['parent_pdg_id']
        trajectories_ancestor_traj_ids = event_products['ancestor_traj_id_map']
        trajectories_ancestor_pdg_ids = event_products['ancestor_pdg_id_map']
        trajectories_ancestor_levels = event_products['ancestor_level_map']
        
        segments_traj_ids = segments['traj_id']
        segments_vertex_ids = segments['vertex_id']
        segments_segment_ids = segments['segment_id']

        segments_event = arrakis_segment_charge['event_id']
        segments_particle = arrakis_segment_charge['particle']
        segments_distance = arrakis_segment_charge['segment_distance']
        segments_topology = arrakis_segment_charge['topology']
        segments_physics = arrakis_segment_charge['physics']
        segments_unique_topology = arrakis_segment_charge['unique_topology']
        segments_fraction = arrakis_segment_charge['segment_fraction']
        segments_vertex = arrakis_segment_charge['vertex']
        segments_tracklette_begin = arrakis_segment_charge['tracklette_begin']
        segments_tracklette_end = arrakis_segment_charge['tracklette_end']
        segments_fragment_begin = arrakis_segment_charge['fragment_begin']
        segments_fragment_end = arrakis_segment_charge['fragment_end']
        segments_shower_begin = arrakis_segment_charge['shower_begin']

        charge_segment_ids = charge_back_track['segment_id'].astype(int)
        charge_segment_fraction = charge_back_track['fraction']
        charge_segment_fraction_mask = (charge_segment_fraction == 0)
        charge_segment_ids[charge_segment_fraction_mask] = -1

        """Iterate over the individual (hits)"""
        for ii, particle in enumerate(segments_particle):
            segment_event = segments_event[ii]
            segment_particle = segments_particle[ii]
            segment_distance = segments_distance[ii]
            segment_topology = segments_topology[ii]
            segment_physics = segments_physics[ii]
            segment_unique_topology = segments_unique_topology[ii]
            segment_fraction = segments_fraction[ii]
            segment_vertex = segments_vertex[ii]
            segment_tracklette_begin = segments_tracklette_begin[ii]
            segment_tracklette_end = segments_tracklette_end[ii]
            segment_fragment_begin = segments_fragment_begin[ii]
            segment_fragment_end = segments_fragment_end[ii]
            segment_shower_begin = segments_shower_begin[ii]
            """Make a cut on segment_influence distance"""
            """
            We want to limit the influence of segments to labeling by how far away they are
            For example, if we have a delta electron coming from a muon, it is very likely
            that induction from the muon will dominate the fraction of energy contained
            in the delta hits.
            """
            """First check to see if any segments remain after the cut"""
            if sum(segment_distance <= self.segment_influence_cut):
                segment_fraction[(segment_distance > self.segment_influence_cut)] = 0
            
            """Then, give priority to tracks"""
            if Topology.Track.value in segment_topology:
                segment_fraction[(segment_topology != Topology.Track.value)] = 0
            """Then, to showers"""
            if Topology.Shower.value in segment_topology:
                segment_fraction[(segment_topology != Topology.Shower.value)] = 0
            segment_distance[(segment_fraction == 0.0)] = 10e10

            if self.constraint_mode == "max_fraction":
                constraint_mask = np.argmax(segment_fraction)
            else:
                constraint_mask = np.argmin(segment_distance)

            arrakis_charge['event_id'][ii] = segment_event
            arrakis_charge['topology'][ii] = segment_topology[constraint_mask]
            arrakis_charge['particle'][ii] = segment_particle[constraint_mask]
            arrakis_charge['physics'][ii] = segment_physics[constraint_mask]
            arrakis_charge['unique_topology'][ii] = segment_unique_topology[constraint_mask]
            arrakis_charge['vertex'][ii] = int(np.any(segment_vertex))
            arrakis_charge['tracklette_begin'][ii] = int(np.any(segment_tracklette_begin))
            arrakis_charge['tracklette_end'][ii] = int(np.any(segment_tracklette_end))
            arrakis_charge['fragment_begin'][ii] = int(np.any(segment_fragment_begin))
            arrakis_charge['fragment_end'][ii] = int(np.any(segment_fragment_end))
            arrakis_charge['shower_begin'][ii] = int(np.any(segment_shower_begin))
            
            """Check for undefined values"""
            if arrakis_charge['topology'][ii] == -1:
                """Replace undefined values"""
                arrakis_charge['topology'][ii] = self.replace_undefined
                arrakis_charge['physics'][ii] = self.replace_undefined
                
                """Log the undefined information to the arrakis file"""
                segment_index = np.where(
                    segments_segment_ids == charge_segment_ids[ii][constraint_mask]
                )
                traj_id = segments_traj_ids[segment_index]
                vertex_id = segments_vertex_ids[segment_index]
                traj_index = np.where(
                    (trajectories_traj_ids == traj_id) &
                    (trajectories_vertex_ids == vertex_id)
                )
                parent_id = trajectories_parent_ids[traj_index]
                pdg_id = trajectories_pdg_ids[traj_index]
                parent_pdg_id = trajectories_parent_pdg_ids[traj_index]
                ancestor_id = trajectories_ancestor_traj_ids[(traj_id[0], vertex_id[0])]
                ancestor_pdg_id = trajectories_ancestor_pdg_ids[(traj_id[0], vertex_id[0])]
                ancestor_level = trajectories_ancestor_levels[(traj_id[0], vertex_id[0])]
                start_process = trajectories_start_process[traj_index]
                start_subprocess = trajectories_start_subprocess[traj_index]
                end_process = trajectories_end_process[traj_index]
                end_subprocess = trajectories_end_subprocess[traj_index]
                parent_index = np.where(
                    (trajectories_traj_ids == parent_id) &
                    (trajectories_vertex_ids == vertex_id)
                )
                if parent_id != -1:
                    parent_start_process = trajectories_start_process[parent_index]
                    parent_start_subprocess = trajectories_start_subprocess[parent_index]
                    parent_end_process = trajectories_end_process[parent_index]
                    parent_end_subprocess = trajectories_end_subprocess[parent_index]
                else:
                    parent_start_process = -1
                    parent_start_subprocess = -1
                    parent_end_process = -1
                    parent_end_subprocess = -1

                undefined_data = np.array([(
                    event,
                    vertex_id,
                    traj_id,
                    parent_id,
                    pdg_id,
                    parent_pdg_id,
                    ancestor_id,
                    ancestor_pdg_id,
                    ancestor_level,
                    start_process,
                    start_subprocess,
                    end_process,
                    end_subprocess,
                    parent_start_process,
                    parent_start_subprocess,
                    parent_end_process,
                    parent_end_subprocess)],
                    dtype=undefined_data_type
                )
                
                """Add the undefined data to the event products"""
                event_products['undefined'].append(undefined_data)
                
        """Write changes to arrakis_file"""
        arrakis_file['charge/calib_final_hits/data'][event_indices['charge']] = arrakis_charge



"""
Script for generating hyperparameter config files from a base config
"""
import os
import numpy as np
import argparse
import csv
import glob

from arrakis_nd.utils.logger import Logger
from arrakis_nd.utils.config import ConfigParser


def run():
    """
    This program splits up an Arrakis config over several config
    files by distributing the flow_files into separate jobs.  This
    allows Arrakis to run on multiple files at once, which greatly
    speeds up runtime.
    """
    parser = argparse.ArgumentParser(
        prog='ArrakisND Config Generator',
        description='This program splits up an Arrakis config over several config \
            files by distributing the flow_files into separate jobs.  This \
            allows Arrakis to run on multiple files at once, which greatly \
            speeds up runtime.',
        epilog='...'
    )
    parser.add_argument(
        'config_file',
        metavar='<str>.yml',
        type=str,
        help='config file specification for a module.'
    )
    parser.add_argument(
        '-arrakis_config_location',
        dest='arrakis_config_location',
        default='/local_arrakis',
        help='location for the local arrakis directory.'
    )
    parser.add_argument(
        '-number_of_processes',
        dest='number_of_processes',
        default=1,
        help='number of processes for arrakis runs.'
    )

    """Set up logger"""
    logger = Logger('arrakis_generator')

    """Parse command line arguments"""
    args = parser.parse_args()
    config_file = args.config_file
    arrakis_config_location = args.arrakis_config_location
    number_of_processes = int(args.number_of_processes)

    """Pull in config file"""
    config_parser = ConfigParser(config_file)
    config = config_parser.data

    arrakis_dict = config['arrakis_nd']

    """Check for parameters"""
    if 'flow_folder' not in arrakis_dict.keys():
        logger.error('flow_folder not specified in config!')
    if 'flow_files' not in arrakis_dict.keys():
        logger.error('flow_files not specified in config!')

    flow_folder = arrakis_dict['flow_folder']
    flow_files = arrakis_dict["flow_files"]

    """Check for arrakis folder"""
    if 'arrakis_folder' not in arrakis_dict.keys():
        logger.warn('arrakis_folder not specified in config! setting to "/local_scratch"')
        arrakis_dict['arrakis_folder'] = '/local_scratch'
    arrakis_folder = arrakis_dict['arrakis_folder'].replace('flow', 'arrakis').replace('FLOW', 'ARRAKIS')

    """Check that flow folder exists"""
    if not os.path.isdir(flow_folder):
        logger.error(f'specified flow_folder {flow_folder} does not exist!')

    """Check that flow folder has a '/' at the end"""
    if flow_folder[-1] != '/':
        flow_folder += '/'

    """Check that arrakis folder has a '/' at the end"""
    if arrakis_folder[-1] != '/':
        arrakis_folder += '/'

    """Check that arrakis folder exists"""
    if not os.path.isdir(arrakis_folder):
        os.makedirs(arrakis_folder)

    if isinstance(arrakis_dict["flow_files"], list):
        """
        If the flow_files parameter is a list, look through
        the list and make sure each specified file actually exists
        in the flow_folder.
        """
        flow_files = [
            input_file for input_file in arrakis_dict["flow_files"]
            if input_file not in arrakis_dict["skip_files"]
        ]
        for flow_file in flow_files:
            if not os.path.isfile(flow_file):
                logger.error(
                    f"specified file {flow_file} does not exist in directory {flow_folder}!"
                )
    elif isinstance(arrakis_dict["flow_files"], str):
        """
        If the flow_files parameter is a string, check if its
        the phrase 'all', and if so, recursively grab all h5
        files in the flow_folder.

        Otherwise, assume that the flow_files parameter is a
        file extension, and search recursively for all files
        with that extension.
        """
        if arrakis_dict["flow_files"] == "all":
            logger.info(
                f"searching {flow_folder} recursively for all .h5 FLOW files."
            )
            flow_files = [
                os.path.basename(input_file) for input_file in glob.glob(
                    f"{flow_folder}*.h5", recursive=True
                )
                if 'FLOW' in input_file and input_file not in arrakis_dict["skip_files"]
            ]
        else:
            try:
                logger.info(
                    f'searching {flow_folder} recursively for all {arrakis_dict["flow_files"]} files.'
                )
                flow_files = [
                    os.path.basename(input_file) for input_file in glob.glob(
                        f'{flow_folder}/*.{arrakis_dict["flow_files"]}',
                        recursive=True,
                    )
                    if input_file not in arrakis_dict["skip_files"]
                ]
            except Exception as exception:
                logger.error(
                    f'specified "flow_files" parameter: {arrakis_dict["flow_files"]} incompatible!'
                    + f" exception: {exception}"
                )
    else:
        logger.error(
            f'specified "flow_files" parameter: {arrakis_dict["flow_files"]} incompatible!'
        )
    if len(flow_files) == 0:
        logger.error(f"found no flow files in {flow_folder}!")
    logger.info(f"found {len(flow_files)} flow files")

    """Split up flow files among the jobs"""
    file_chunk_size = len(flow_files) // number_of_processes
    larger_file_chunk_size = len(flow_files) % number_of_processes
    logger.info(f"file_chunk_size is {file_chunk_size}")
    logger.info(f"remainder size is {larger_file_chunk_size}")

    file_chunks = []
    start = 0
    for ii in range(number_of_processes):
        end = start + file_chunk_size + (1 if ii < larger_file_chunk_size else 0)
        file_chunks.append(flow_files[start:end])
        start = end
    logger.info(f"split up {len(flow_files)} into {number_of_processes}")

    """Generate the corresponding config files for each job"""
    for ii in range(len(file_chunks)):
        config_iteration = config.copy()
        config_iteration['arrakis_nd']['flow_files'] = file_chunks[ii]
        config_parser.save_config(
            config_iteration,
            f'{os.path.abspath(arrakis_config_location)}/arrakis_config_{ii}.yaml'
        )
    logger.info(f"saving split configs into {arrakis_config_location}")
    logger.info("create_arrakis_runs job complete. closing out")


if __name__ == "__main__":
    run()



"""
ArrakisND main program
"""
import argparse

from arrakis_nd.utils.logger import Logger
from arrakis_nd.utils.display.arrakis_display import ArrakisDisplay


def run():
    """
    This program runs the Arrakis display and generates a 
    dash server instance that can be run on NERSC ...
    """

    """Set up command line arguments"""
    parser = argparse.ArgumentParser(
        prog="Arrakis Display Runner",
        description="This program runs the Arrakis display and generates a \
            dash server instance that can be run on NERSC ... ",
        epilog="...",
    )

    """Parse command line arguments"""
    args = parser.parse_args()
    
    """Create the Arrakis instance"""
    try:
        arrakis = ArrakisDisplay()
    except Exception as e:
        raise RuntimeError(f"failed to construct Arrakis object: {e}")

    """Run Arrakis"""
    arrakis.run_app()


if __name__ == "__main__":
    run()
